#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the TAR Software class."""

from __future__ import absolute_import
import unittest
import os
import sys
from mock import mock_open, patch, MagicMock as Mock

import six
from six.moves import range

from DIRAC import S_OK, S_ERROR
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds, assertDiracSucceedsWith_equals, assertMockCalls
from Tests.Utilities.FileUtils import FileUtil

__RCSID__ = "$Id$"

BUILTIN = 'builtins' if six.PY3 else '__builtin__'


class TestTARsoft(unittest.TestCase):  # pylint: disable=too-many-public-methods
  """Tests the base functionality of the class."""

  def setUp(self):
    # Mock away DataManager import
    dataman_import_mock = Mock()
    dataman_import_mock.getFile.return_value = None
    self.dm_backup = sys.modules.get('DIRAC.DataManagementSystem.Client.DataManager', -1)
    sys.modules['DIRAC.DataManagementSystem.Client.DataManager'] = dataman_import_mock
    self.dataman_mock = Mock()

  def tearDown(self):
    if self.dm_backup != -1:
      sys.modules['DIRAC.DataManagementSystem.Client.DataManager'] = self.dm_backup
    else:
      del sys.modules['DIRAC.DataManagementSystem.Client.DataManager']

  def test_createlock(self):
    from ILCDIRAC.Core.Utilities.TARsoft import createLock
    file_contents = [[]]
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mo.side_effect = (h for h in handles)
      assertDiracSucceeds(createLock('testfile.py'), self)
      mo.assert_called_once_with('testfile.py', 'w')
      expected = [['Locking this directory\n']]
      self.assertEqual(len(file_contents), len(expected))
      for (index, handle) in enumerate(handles):
        cur_handle = handle.__enter__()
        self.assertEqual(len(expected[index]), handle.__enter__.return_value.write.call_count)
        for entry in expected[index]:
          cur_handle.write.assert_any_call(entry)

  def test_createlock_ioerr(self):
    from ILCDIRAC.Core.Utilities.TARsoft import createLock
    with patch('%s.open' % MODULE_NAME, new=Mock(side_effect=IOError('some_test_open_error')), create=True):
      assertDiracFailsWith(createLock('myfile.txt'),
                            'not allowed to write here: ioerror some_test_open_error', self)

  def test_check_lock_age_no_lock(self):
    from ILCDIRAC.Core.Utilities.TARsoft import checkLockAge
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)):
      result = checkLockAge('mylockfile.txt')
      assertDiracSucceedsWith_equals(result, False, self)

  def test_check_lock_age_stat_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import checkLockAge
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.stat' % MODULE_NAME, new=Mock(side_effect=OSError('some_test_stat_error'))), \
         patch('%s.wasteCPUCycles' % MODULE_NAME, new=Mock(return_value=S_OK())):
      result = checkLockAge('mylockfile.txt')
      assertDiracSucceedsWith_equals(result, False, self)

  def test_check_lock_age_clear_lock_success(self):
    import time
    from ILCDIRAC.Core.Utilities.TARsoft import checkLockAge
    stat_mock = Mock()
    stat_mock.st_atime = time.time() - 1801  # lastTouch old enough
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.stat' % MODULE_NAME, new=Mock(return_value=stat_mock)), \
         patch('%s.wasteCPUCycles' % MODULE_NAME, new=Mock(side_effect=[S_ERROR(), S_OK(), S_OK()])), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(side_effect=[S_ERROR(), S_OK(True)])):
      result = checkLockAge('mylockfile.txt')
      assertDiracSucceedsWith_equals(result, True, self)

  def test_check_lock_age_timeout(self):
    import time
    from ILCDIRAC.Core.Utilities.TARsoft import checkLockAge
    stat_mock = Mock()
    stat_mock.st_atime = time.time()
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.stat' % MODULE_NAME, new=Mock(return_value=stat_mock)), \
         patch('%s.wasteCPUCycles' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=S_OK(True))):
      result = checkLockAge('mylockfile.txt')
      assertDiracFailsWith(result, 'buggy lock, removed: True', self)

  def test_clear_lock(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clearLock
    with patch('%s.os.unlink' % MODULE_NAME, new=Mock(return_value=True)):
      assertDiracSucceeds(clearLock('mylock'), self)

  def test_clear_lock_oserr(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clearLock
    with patch('%s.os.unlink' % MODULE_NAME, new=Mock(side_effect=OSError('test_unlink_some_err'))):
      assertDiracFailsWith(clearLock('mylock'), 'failed to clear lock: test_unlink_some_err', self)

  def test_delete_old(self):
    from ILCDIRAC.Core.Utilities.TARsoft import deleteOld
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)):
      assertDiracSucceeds(deleteOld('testfile'), self)

  def test_delete_old_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import deleteOld
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, False])) as exists_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock:
      assertDiracSucceeds(deleteOld('testFile_deleteMe.txt'), self)
      file_to_check = 'testFile_deleteMe.txt'
      remove_mock.assert_called_once_with(file_to_check)
      isdir_mock.assert_called_once_with(file_to_check)
      exists_mock.assert_called_with(file_to_check)

  def test_delete_old_file_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import deleteOld
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, True])) as exists_mock, \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False)) as isdir_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=OSError('test_remove_cannot_error'))) as remove_mock, \
            patch('%s.LOG.error' % MODULE_NAME) as err_mock:
      assertDiracSucceeds(deleteOld('testFile_deleteMe2.txt'), self)
      file_to_check = 'testFile_deleteMe2.txt'
      remove_mock.assert_called_once_with(file_to_check)
      isdir_mock.assert_called_once_with(file_to_check)
      exists_mock.assert_called_with(file_to_check)
      assertMockCalls(err_mock, [
          ('Failed deleting testFile_deleteMe2.txt because :', 'OSError test_remove_cannot_error'),
          'Oh Oh, something was not right, the directory testFile_deleteMe2.txt is still here'], self)

  def test_delete_old_dir(self):
    from ILCDIRAC.Core.Utilities.TARsoft import deleteOld
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, False])) as exists_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
         patch('%s.shutil.rmtree' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock:
      assertDiracSucceeds(deleteOld('/delete/this'), self)
      file_to_check = '/delete/this'
      remove_mock.assert_called_once_with(file_to_check)
      isdir_mock.assert_called_once_with(file_to_check)
      exists_mock.assert_called_with(file_to_check)

  def test_delete_old_dir_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import deleteOld
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, True])) as exists_mock, \
        patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
        patch('%s.shutil.rmtree' % MODULE_NAME,
              new=Mock(side_effect=OSError('test_remove_dir_cannot_error'))) as remove_mock, \
        patch('%s.LOG.error' % MODULE_NAME) as err_mock:
      assertDiracSucceeds(deleteOld('/delete/that'), self)
      file_to_check = '/delete/that'
      remove_mock.assert_called_once_with(file_to_check)
      isdir_mock.assert_called_once_with(file_to_check)
      exists_mock.assert_called_with(file_to_check)
      assertMockCalls(err_mock, [
          (('Failed deleting /delete/that because :', "<type 'exceptions.OSError'> test_remove_dir_cannot_error")
           if six.PY2 else
           ('Failed deleting /delete/that because :', "<class 'OSError'> test_remove_dir_cannot_error")),
          'Oh Oh, something was not right, the directory /delete/that is still here',
          ],
          self)

  def test_download_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import downloadFile
    with patch('%s.six.moves.urllib.request.urlretrieve' % MODULE_NAME) as url_mock:
      assertDiracSucceeds(downloadFile('http://my/tarball/url', '/a/basename', '/my/folder/'), self)
      url_mock.assert_called_once_with('http://my/tarball/url//a/basename', 'basename')

  def test_download_file_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import downloadFile
    with patch('%s.six.moves.urllib.request.urlretrieve' % MODULE_NAME,
               new=Mock(side_effect=IOError('test_io_exception'))) as url_mock:
      result = downloadFile('http://my/other_url/', '/other/name/here', '/other_folder/')
      url_mock.assert_called_once_with('http://my/other_url//other/name/here', 'here')
      assertDiracFailsWith(result, 'exception during url retrieve: test_io_exception', self)

  def test_download_file_from_datman(self):
    self.dataman_mock.getFile.return_value = S_OK()
    with patch('%s.DataManager' % MODULE_NAME, new=Mock(return_value=self.dataman_mock)):
      from ILCDIRAC.Core.Utilities.TARsoft import downloadFile
      assertDiracSucceeds(downloadFile('/my/datamanager/directory', 'app/tarball', '/other_folder/'), self)
      self.dataman_mock.getFile.assert_called_once_with('/my/datamanager/directory/app/tarball')

  def test_download_file_from_datman_fails(self):
    self.dataman_mock.getFile.return_value = S_ERROR('datman_test_err')
    with patch('%s.DataManager' % MODULE_NAME, new=Mock(return_value=self.dataman_mock)):
      from ILCDIRAC.Core.Utilities.TARsoft import downloadFile
      assertDiracFailsWith(downloadFile('/unavailable/datman/dir/', 'mytestappv2/tarball',
                                          '/other_folder/'), 'datman_test_err', self)
      self.dataman_mock.getFile.assert_called_once_with('/unavailable/datman/dir/mytestappv2/tarball')

  def test_md5_check(self):
    from ILCDIRAC.Core.Utilities.TARsoft import tarMd5Check
    file_contents = [[b'849utj429foemfi84j92fno;(*ME(FOJN$EO&*R#BNOFMN(OJIm']]
    expected_hash = 'dab9783374461a26e100164747e84e63'  # Precalculated
    handles = FileUtil.getMultipleReadHandles(file_contents, isBytes=True)
    with patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mo.side_effect = (h for h in handles)
      result = tarMd5Check('/my/app/tarball.tgz', expected_hash)
      mo.assert_called_once_with('/my/app/tarball.tgz', 'rb')
      assertDiracSucceeds(result, self)

  def test_md5_check_io_err(self):
    from ILCDIRAC.Core.Utilities.TARsoft import tarMd5Check
    with patch('%s.open' % MODULE_NAME, new=Mock(side_effect=IOError('myerr')), create=True), \
            patch('%s.LOG.warn' % MODULE_NAME, new=Mock()) as log_mock:
      result = tarMd5Check('/my/app/tarball.tgz', '138974')
      assertDiracSucceeds(result, self)
      log_mock.assert_called_once_with('Failed to get tar ball md5, try without')

  def test_md5_check_hash_wrong(self):
    from ILCDIRAC.Core.Utilities.TARsoft import tarMd5Check
    file_contents = [[b'2984jt4gomrfg8924jgnm',
                      b'1938jhfo9coiemc0m90pn@O*E&HQRF(*IONU)',
                      b')(DCIFUJE*FIUEqf9oi1mnf)']]
    expected_hash = '0981u3jr9831rkjopk,f90381'
    handles = FileUtil.getMultipleReadHandles(file_contents, isBytes=True)
    with patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mo.side_effect = (h for h in handles)
      result = tarMd5Check('/my/app/tarball.tgz', expected_hash)
      mo.assert_called_once_with('/my/app/tarball.tgz', 'rb')
      assertDiracFailsWith(result, 'hash does not correspond', self)

  def test_install_deps(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installDependencies
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'myappname1', 'version': '203.0'}, {'app': 'myappname2', 'version': '138.1'}])) as dep_mock, \
         patch('%s.installInAnyArea' % MODULE_NAME) as install_mock:
      result = installDependencies(('AppName', 'appvers'), 'myconf', 'myareas')
      assertDiracSucceeds(result, self)
      assertMockCalls(install_mock, [('myareas', ['myappname1', '203.0'], 'myconf'),
                                       ('myareas', ['myappname2', '138.1'], 'myconf')],
                       self, only_these_calls=False)
      dep_mock.assert_called_once_with('myconf', 'appname', 'appvers')

  def test_install_deps_nodeps(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installDependencies
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[])) as dep_mock, \
         patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(return_value=S_OK())) as install_mock:
      result = installDependencies(('AppName', 'appvers'), 'myconf', 'myareas')
      assertDiracSucceeds(result, self)
      self.assertFalse(install_mock.called)
      dep_mock.assert_called_once_with('myconf', 'appname', 'appvers')

  def test_install_deps_installation_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installDependencies
    with patch('%s.resolveDeps' % MODULE_NAME, new=Mock(return_value=[{'app': 'myappname1', 'version': '203.0'}, {'app': 'myappname2', 'version': '138.1'}])) as dep_mock, \
         patch('%s.installInAnyArea' % MODULE_NAME, new=Mock(side_effect=[S_OK(), S_ERROR()])) as install_mock:
      result = installDependencies(('AppName', 'appvers'), 'myconf', 'myareas')
      assertDiracFailsWith(result, "failed to install dependency: ['myappname2', '138.1']", self)
      assertMockCalls(install_mock, [('myareas', ['myappname1', '203.0'], 'myconf'),
                                       ('myareas', ['myappname2', '138.1'], 'myconf')], self)
      dep_mock.assert_called_once_with('myconf', 'appname', 'appvers')

  def test_install_in_any_area(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installInAnyArea
    with patch('%s.installSinglePackage' % MODULE_NAME, new=Mock(side_effect=[S_ERROR('not here'), S_ERROR('neither here'), S_OK('Bingo')])) as install_mock:
      assertDiracSucceeds(installInAnyArea(['area1', 'otherarea', 'thisonesurelyworks'],
                                             'myapplication2', 'jobconf'), self)
      assertMockCalls(install_mock, [('myapplication2', 'jobconf', 'area1'),
                                       ('myapplication2', 'jobconf', 'otherarea'),
                                       ('myapplication2', 'jobconf', 'thisonesurelyworks')], self)

  def test_install_in_any_area_fail_all(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installInAnyArea
    with patch('%s.installSinglePackage' % MODULE_NAME, new=Mock(return_value=S_ERROR('not here'))) as install_mock:
      assertDiracFailsWith(installInAnyArea(['area1', 'otherarea', 'thisonesurelyworks'],
                                              'myapplication3', 'jobconf'), 'failed to install software',
                            self)
      assertMockCalls(install_mock, [('myapplication3', 'jobconf', 'area1'),
                                       ('myapplication3', 'jobconf', 'otherarea'),
                                       ('myapplication3', 'jobconf', 'thisonesurelyworks')], self)

  def test_install_single_package(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installSinglePackage
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='mycwd')), \
         patch('%s.installPackage' % MODULE_NAME, new=Mock(return_value=S_OK(''))) as install_mock:
      result = installSinglePackage(('appLication', 'v201'), 'jobConfiguration', 'Area51')
      assertDiracSucceeds(result, self)
      install_mock.assert_called_once_with(('appLication', 'v201'), 'jobConfiguration', 'Area51', 'mycwd')

  def test_install_single_package_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installSinglePackage
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='mycwd')), \
         patch('%s.installPackage' % MODULE_NAME, new=Mock(return_value=S_ERROR(''))) as install_mock:
      result = installSinglePackage(('appLication', 'v202'), 'jobConfiguration', 'Area51')
      assertDiracFailsWith(result, 'failed to install here', self)
      install_mock.assert_called_once_with(('appLication', 'v202'), 'jobConfiguration', 'Area51', 'mycwd')

  def test_get_tarball_location(self):
    from ILCDIRAC.Core.Utilities.TARsoft import getTarBallLocation
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(side_effect=['tarball_loc', True, 'mymd5sum', 'mytarURL'])) as getval_mock:
      result = getTarBallLocation(('complicated_app', 'v201'), 'complex_config', 'dummy_area')
      assertDiracSucceedsWith_equals(result, ['tarball_loc', 'mytarURL', True, 'mymd5sum'], self)
      assertMockCalls(getval_mock, [('/AvailableTarBalls/complex_config/complicated_app/v201/TarBall', ''),
                                      ('/AvailableTarBalls/complex_config/complicated_app/v201/Overwrite', False),
                                      ('/AvailableTarBalls/complex_config/complicated_app/v201/Md5Sum', ''),
                                      ('/AvailableTarBalls/complex_config/complicated_app/TarBallURL', '')],
                       self)

  def test_get_tarball_location_noapptar(self):
    from ILCDIRAC.Core.Utilities.TARsoft import getTarBallLocation
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(side_effect=['', True, 'mymd5sum', 'mytarURL'])):
      result = getTarBallLocation(('complicated_app', 'v201'), 'config', 'dummy_area')
      assertDiracFailsWith(result, 'could not find tar ball for complicated_app v201', self)

  def test_get_tarball_location_nourl(self):
    from ILCDIRAC.Core.Utilities.TARsoft import getTarBallLocation
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(side_effect=['tarball_loc', True, 'mymd5sum', ''])):
      result = getTarBallLocation(('complicated_app', 'v201'), 'config', 'dummy_area')
      assertDiracFailsWith(result, 'could not find tarballurl in cs', self)

  def test_check(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    # Different treatment of with open() and open().read().....
    file_contents = [['HelloWorld -',
                      'mychecksum md5_checksum.md5',
                      'ef7493baf5582b41cd4fcaa25124c3d9 ./appfile1.txt',
                      '6fe635186392d43f0ae5728e41a92f96 appfile2.ppt',
                      'e10dab217514cf7f2166d87fb5d04d5c ./myapp/importantfile.bin',
                      'a51999cbc7ee434ee95b33f1d18c7210 otherdir/main.py',
                      'abc libstdc++.so', 'def myapp/libgcc_s.so.1',
                      'ignoreme ./other_dir/libc-2.5']]
    other_file_contents = [b'appfile1r0984u3jriumfilf42890tjr742tu',
                           b'appfile29031i4rt498jnyfouf908j248f4298fn24iuyf',
                           b'importf90ui4j9rf41f09j14fiun41 fp1,cmic13',
                           b'MAIN()@KJR(*@KRE)+@MOUFRIN@FR*YB^ B* @HE)J @E( HG!V!UYNEJ)(!))))']
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mysideeff = [handles[0]]
      for i in range(0, len(other_file_contents)):
        other_read_mock = Mock()
        other_read_mock.read.return_value = other_file_contents[i]
        mysideeff.append(other_read_mock)
      mo.side_effect = tuple(mysideeff)
      result = check(('appname', 'version'), 'mytestarea398', ['/my/basefolder', 'res_from_install[1]'])
      expected_open_calls = [('/my/basefolder/md5_checksum.md5', 'r'),
                             ('/my/basefolder/appfile1.txt', 'rb'),
                             ('/my/basefolder/appfile2.ppt', 'rb'),
                             ('/my/basefolder/myapp/importantfile.bin', 'rb'),
                             ('/my/basefolder/otherdir/main.py', 'rb')]
      expected = [[]]
      FileUtil.checkFileInteractions(self, mo, expected_open_calls, expected, handles)
      assertDiracSucceedsWith_equals(result, ['/my/basefolder'], self)
      chdir_mock.assert_called_once_with('mytestarea398')
      isfile_mock.assert_called_once_with('/my/basefolder')
      assertMockCalls(exists_mock, ['/my/basefolder/md5_checksum.md5', '/my/basefolder/appfile1.txt',
                                      '/my/basefolder/appfile2.ppt', '/my/basefolder/myapp/importantfile.bin',
                                      '/my/basefolder/otherdir/main.py'], self)

  def test_check_lcsim_jar(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    with patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=True)) as isfile_mock, \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock:
      result = check(('appname', 'version'), 'deep/area', ['mytestbasefolder', 'res_from_install[1]'])
      assertDiracSucceedsWith_equals(result, ['mytestbasefolder'], self)
      isfile_mock.assert_called_once_with('mytestbasefolder')
      chdir_mock.assert_called_once_with('deep/area')

  def test_check_corrupt_checksum(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    # Different treatment of with open() and open().read().....
    file_contents = [['HelloWorld -',
                      'mychecksum md5_checksum.md5',
                      'ef7493baf5582b41cd4fcaa25124c3d9 ./appfile1.txt',
                      'CORRUPT_CHECKSUM appfile2.ppt',
                      'e10dab217514cf7f2166d87fb5d04d5c ./myapp/importantfile.bin',
                      'a51999cbc7ee434ee95b33f1d18c7210 otherdir/main.py',
                      'abc libstdc++.so',
                      'def myapp/libgcc_s.so.1',
                      'ignoreme ./other_dir/libc-2.5']]
    other_file_contents = [b'appfile1r0984u3jriumfilf42890tjr742tu',
                           b'appfile29031i4rt498jnyfouf908j248f4298fn24iuyf',
                           b'importf90ui4j9rf41f09j14fiun41 fp1,cmic13',
                           b'MAIN()@KJR(*@KRE)+@MOUFRIN@FR*YB^ B* @HE)J @E( HG!V!UYNEJ)(!))))']
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
         patch('%s.open' % BUILTIN, mock_open(), create=True) as mo:
      mysideeff = [handles[0]]
      for content in other_file_contents:
        other_read_mock = Mock()
        other_read_mock.read.return_value = content
        mysideeff.append(other_read_mock)
      mo.side_effect = tuple(mysideeff)
      result = check(('appname', 'version'), 'mytestarea398', ['/my/otherfolder', 'res_from_install[1]'])
      assertDiracFailsWith(result, 'corrupted install: file /my/otherfolder/appfile2.ppt', self)
      chdir_mock.assert_called_once_with('mytestarea398')
      isfile_mock.assert_called_once_with('/my/otherfolder')
      assertMockCalls(exists_mock, ['/my/otherfolder/md5_checksum.md5', '/my/otherfolder/appfile1.txt',
                                      '/my/otherfolder/appfile2.ppt'], self)

  def test_check_no_checksum_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
        patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
            patch('%s.LOG.warn' % MODULE_NAME) as warn_mock:
      result = check(('appname', 'version'), 'mytestarea398', ['/base/', 'res_from_install[1]'])
      assertDiracSucceedsWith_equals(result, ['/base/'], self)
      chdir_mock.assert_called_once_with('mytestarea398')
      isfile_mock.assert_called_once_with('/base/')
      exists_mock.assert_called_once_with('/base/md5_checksum.md5')
      warn_mock.assert_called_once_with('The application does not come with md5 checksum file:',
                                         ('appname', 'version'))

  def test_check_ioerr(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    # Different treatment of with open() and open().read().....
    file_contents = [['ef7493baf5582b41cd4fcaa25124c3d9 ./appfile1.txt',
                        '6fe635186392d43f0ae5728e41a92f96 appfile2.ppt',
                        'e10dab217514cf7f2166d87fb5d04d5c ./myapp/importantfile.bin',
                        'a51999cbc7ee434ee95b33f1d18c7210 otherdir/main.py', 'abc libstdc++.so',
                        'def myapp/libgcc_s.so.1', 'ignoreme ./other_dir/libc-2.5']]
    other_file_contents = [b'appfile1r0984u3jriumfilf42890tjr742tu', IOError('md5_read_err'),
                            b'importf90ui4j9rf41f09j14fiun41 fp1,cmic13',
                            b'MAIN()@KJR(*@KRE)+@MOUFRIN@FR*YB^ B* @HE)J @E( HG!V!UYNEJ)(!))))']
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mysideeff = [handles[0]]
      for other_file_content in other_file_contents:
        other_read_mock = Mock()
        if isinstance(other_file_content, (six.string_types, bytes)):
          other_read_mock.read.side_effect = [other_file_content]
        else:
          other_read_mock.read.side_effect = other_file_content
        mysideeff.append(other_read_mock)
      mo.side_effect = tuple(mysideeff)
      result = check(('appname', 'version'), 'mytestarea398', ['/my/basefolder', 'res_from_install[1]'])
      expected_open_calls = [('/my/basefolder/md5_checksum.md5', 'r'),
                             ('/my/basefolder/appfile1.txt', 'rb'),
                            ('/my/basefolder/appfile2.ppt', 'rb')]
      expected = [[]]
      assertDiracFailsWith(result, 'failed to compute md5 sum', self)
      FileUtil.checkFileInteractions(self, mo, expected_open_calls, expected, handles)
      chdir_mock.assert_called_once_with('mytestarea398')
      isfile_mock.assert_called_once_with('/my/basefolder')
      assertMockCalls(exists_mock, ['/my/basefolder/md5_checksum.md5', '/my/basefolder/appfile1.txt',
                                      '/my/basefolder/appfile2.ppt'], self)

  def test_check_empty_checksum_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    file_contents = [[]]
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
        patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo, \
            patch('%s.LOG.warn' % MODULE_NAME) as warn_mock:
      mo.side_effect = (h for h in handles)
      result = check(('appname', 'version'), 'mytestarea398', ['/my/basefolder', 'res_from_install[1]'])
      expected_open_calls = [('/my/basefolder/md5_checksum.md5', 'r')]
      expected = [[]]
      FileUtil.checkFileInteractions(self, mo, expected_open_calls, expected, handles)
      assertDiracSucceedsWith_equals(result, ['/my/basefolder'], self)
      chdir_mock.assert_called_once_with('mytestarea398')
      self.assertFalse(warn_mock.called)
      isfile_mock.assert_called_once_with('/my/basefolder')
      exists_mock.assert_called_once_with('/my/basefolder/md5_checksum.md5')

  def test_check_missing_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import check
    # Different treatment of with open() and open().read().....
    file_contents = [['HelloWorld -', 'mychecksum md5_checksum.md5',
                       'ef7493baf5582b41cd4fcaa25124c3d9 ./appfile1.txt',
                       '6fe635186392d43f0ae5728e41a92f96 appfile2.ppt',
                       'e10dab217514cf7f2166d87fb5d04d5c ./myapp/importantfile.bin',
                       'a51999cbc7ee434ee95b33f1d18c7210 otherdir/main.py', 'abc libstdc++.so',
                       'def myapp/libgcc_s.so.1', 'ignoreme ./other_dir/libc-2.5']]
    other_file_contents = [b'appfile1r0984u3jriumfilf42890tjr742tu',
                           b'appfile29031i4rt498jnyfouf908j248f4298fn24iuyf',
                           b'importf90ui4j9rf41f09j14fiun41 fp1,cmic13',
                           b'MAIN()@KJR(*@KRE)+@MOUFRIN@FR*YB^ B* @HE)J @E( HG!V!UYNEJ)(!))))']
    handles = FileUtil.getMultipleReadHandles(file_contents)
    with patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(return_value=False)) as isfile_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[True, True, False])) as exists_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(), create=True) as mo:
      mysideeff = [handles[0]]
      for other_file_content in other_file_contents:
        other_read_mock = Mock()
        other_read_mock.read.return_value = other_file_content
        mysideeff.append(other_read_mock)
      mo.side_effect = tuple(mysideeff)
      result = check(('appname', 'version'), 'mytestarea398', ['/hidden/dir', 'res_from_install[1]'])
      expected_open_calls = [('/hidden/dir/md5_checksum.md5', 'r'), ('/hidden/dir/appfile1.txt', 'rb')]
      expected = [[]]
      FileUtil.checkFileInteractions(self, mo, expected_open_calls, expected, handles)
      assertDiracFailsWith(result, 'incomplete install: the file /hidden/dir/appfile2.ppt is missing', self)
      chdir_mock.assert_called_once_with('mytestarea398')
      isfile_mock.assert_called_once_with('/hidden/dir')
      assertMockCalls(exists_mock, ['/hidden/dir/md5_checksum.md5',
                                    '/hidden/dir/appfile1.txt',
                                    '/hidden/dir/appfile2.ppt'], self)

  def test_configure_slic(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {}, True), \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(side_effect=[['myslicpaths'], ['mylcddpaths'], ['myxercespaths']])) as listdir_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock:
      result = configure(('SLIC', 'version1000'), 'configurationArea', ['ILCApplication3002'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      exists_mock.assert_called_once_with('/cool/dir/ILCApplication3002/packages/xerces/')
      assertMockCalls(listdir_mock, ['/cool/dir/ILCApplication3002/packages/slic/',
                                       '/cool/dir/ILCApplication3002/packages/lcdd/',
                                       '/cool/dir/ILCApplication3002/packages/xerces/'], self)
      assertEqualsImproved(os.environ, {'SLIC_DIR': '/cool/dir/ILCApplication3002',
                                          'SLIC_VERSION': 'myslicpaths', 'XERCES_VERSION': 'myxercespaths',
                                          'LCDD_VERSION': 'mylcddpaths'}, self)

  def test_configure_slic_emptydirs(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=False))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {}, True), \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=(['']))) as listdir_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock:
      result = configure(('SLIC', 'version1000'), 'configurationArea', ['ILCApplication3002'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      libc_mock.assert_called_once_with('/cool/dir/ILCApplication3002/LDLibs')
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      self.assertFalse(addfolder_mock.called)
      exists_mock.assert_called_once_with('/cool/dir/ILCApplication3002/packages/xerces/')
      assertMockCalls(listdir_mock, ['/cool/dir/ILCApplication3002/packages/slic/',
                                       '/cool/dir/ILCApplication3002/packages/lcdd/'], self)
      assertEqualsImproved(os.environ, {'SLIC_DIR': '/cool/dir/ILCApplication3002'}, self)

  def test_configure_slic_oserr(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {}, True), \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(side_effect=[['myslicpaths'], OSError('test_os_listdir_err')])) as listdir_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock:
      result = configure(('SLIC', 'version1000'), 'configurationArea', ['ILCApplication3002'])
      assertDiracSucceeds(result, self)  # FIXME: OSError in configureSlic is ignored... Intentional?
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      self.assertFalse(exists_mock.called)
      assertMockCalls(listdir_mock, ['/cool/dir/ILCApplication3002/packages/slic/',
                                       '/cool/dir/ILCApplication3002/packages/lcdd/'], self)
      assertEqualsImproved(os.environ, {'SLIC_DIR': '/cool/dir/ILCApplication3002'}, self)

  def test_configure_unknownapp(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock:
      result = configure(('myILCApplicationTesttest123', 'version1000'), 'configurationArea',
                          ['ILCApplication3002'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')

  def test_configure_root(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/current/working/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {'PATH': 'mypath', 'PYTHONPATH': 'pythonic_path'}, True):
      result = configure(('ROOT', 'v9084u'), 'configurationArea', ['myres'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/current/working/dir/myres/LDLibs', '/current/working/dir/myres/lib'],
                       self)
      isdir_mock.assert_called_once_with('/current/working/dir/myres/lib')
      assertMockCalls(addfolder_mock, ['/current/working/dir/myres/lib'] * 2,
                       self)
      assertEqualsImproved(os.environ, {'PATH': '/current/working/dir/myres/bin:mypath',
                                          'PYTHONPATH': '/current/working/dir/myres/lib:pythonic_path',
                                          'ROOTSYS': '/current/working/dir/myres'}, self)

  def test_configure_java(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/java/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {'PATH': 'mypath'}, True):
      result = configure(('Java', 'v9084u'), 'configurationArea', ['res'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/java/res/LDLibs', '/java/res/lib'], self)
      isdir_mock.assert_called_once_with('/java/res/lib')
      assertMockCalls(addfolder_mock, ['/java/res/lib'] * 2, self)
      assertEqualsImproved(os.environ, {'PATH': '/java/res/bin:mypath'}, self)

  def test_configure_lcio(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/lcio/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch.dict(os.environ, {'PATH': 'PaTH_LCIO'}, True), \
         patch('%s.subprocess.check_call' % MODULE_NAME, new=Mock(return_value=0)) as subproc_mock:
      result = configure(('LCIO', 'v9084u'), 'configurationArea', ['retval'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/my/lcio/retval/LDLibs', '/my/lcio/retval/lib'], self)
      isdir_mock.assert_called_once_with('/my/lcio/retval/lib')
      addfolder_mock.assert_called_once_with('/my/lcio/retval/lib')
      subproc_mock.assert_called_once_with(['java', '-Xmx1536m', '-Xms256m', '-version'])
      assertEqualsImproved(os.environ, {'LCIO': '/my/lcio/retval',
                                          'PATH': '/my/lcio/retval/bin:PaTH_LCIO'}, self)

  def test_configure_lcio_javafails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch('%s.subprocess.check_call' % MODULE_NAME, new=Mock(return_value=1)) as subproc_mock:
      result = configure(('LCIO', 'version1000'), 'configurationArea', ['ILCApplication3002'])
      assertDiracFailsWith(result, 'something is wrong with java', self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      subproc_mock.assert_called_once_with(['java', '-Xmx1536m', '-Xms256m', '-version'])

  def test_configure_lcsim(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch('%s.subprocess.check_call' % MODULE_NAME, new=Mock(return_value=1)) as subproc_mock:
      result = configure(('LCSIM', 'v9084u'), 'configurationArea', ['ILCApplication3002'])
      assertDiracFailsWith(result, 'something is wrong with java', self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      subproc_mock.assert_called_once_with(['java', '-Xmx1536m', '-Xms256m', '-version'])

  def test_configure_stdhepcutjava(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True)) as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch('%s.subprocess.check_call' % MODULE_NAME, new=Mock(return_value=0)) as subproc_mock:
      result = configure(('StdHEPCutJava', 'v9084u'), 'configurationArea', ['ILCApplication3002'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      subproc_mock.assert_called_once_with(['java', '-Xmx1536m', '-Xms256m', '-version'])

  def test_configure_stdhepcutjava_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import configure
    import subprocess
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cool/dir/')), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
         patch('%s.removeLibc' % MODULE_NAME, new=Mock(return_value=True)) as libc_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(return_value=True))as isdir_mock, \
         patch('%s.addFolderToLdLibraryPath' % MODULE_NAME, new=Mock(return_value=True)) as addfolder_mock, \
         patch('%s.subprocess.check_call' % MODULE_NAME, new=Mock(side_effect=subprocess.CalledProcessError(1, 'testcmd'))) as subproc_mock:
      result = configure(('StdHEPCutJava', 'v9084u'), 'configurationArea', ['ILCApplication3002'])
      assertDiracFailsWith(result, 'java was not found on this machine, cannot proceed', self)
      chdir_mock.assert_called_once_with('configurationArea')
      assertMockCalls(libc_mock, ['/cool/dir/ILCApplication3002/LDLibs', '/cool/dir/ILCApplication3002/lib'],
                       self)
      isdir_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      addfolder_mock.assert_called_once_with('/cool/dir/ILCApplication3002/lib')
      subproc_mock.assert_called_once_with(['java', '-Xmx1536m', '-Xms256m', '-version'])

  def test_clean(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clean
    with patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
         patch('%s.os.unlink' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/current/dir/my')):
      result = clean('mytestarea51', ['res_from_check[0]', 'myapptarball.tgz'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('mytestarea51')
      exists_mock.assert_called_once_with('/current/dir/my/myapptarball.tgz')
      remove_mock.assert_called_once_with('myapptarball.tgz')

  def test_clean_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clean
    with patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.os.unlink' % MODULE_NAME, new=Mock(side_effect=OSError('_unlink_some_err'))) as remove_mock, \
        patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/current/dir/my')), \
            patch('%s.LOG.error' % MODULE_NAME) as err_mock:
      result = clean('mytestarea51', ['res_from_check[0]', 'myapptarball.tgz'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('mytestarea51')
      exists_mock.assert_called_once_with('/current/dir/my/myapptarball.tgz')
      remove_mock.assert_called_once_with('myapptarball.tgz')
      err_mock.assert_called_once_with('Could not remove tar ball:', '_unlink_some_err')

  def test_clean_doesnt_exist(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clean
    with patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
        patch('%s.os.unlink' % MODULE_NAME) as remove_mock, \
        patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/current/dir/my')), \
            patch('%s.LOG.error' % MODULE_NAME) as err_mock:
      result = clean('mytestarea51', ['res_from_check[0]', 'myapptarball.tgz'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('mytestarea51')
      exists_mock.assert_called_once_with('/current/dir/my/myapptarball.tgz')
      self.assertFalse(remove_mock.called)
      self.assertFalse(err_mock.called)

  def test_clean_dont_delete_jar(self):
    from ILCDIRAC.Core.Utilities.TARsoft import clean
    with patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)) as exists_mock, \
        patch('%s.os.unlink' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/current/dir/my')), \
            patch('%s.LOG.error' % MODULE_NAME) as err_mock:
      result = clean('mytestarea51', ['res_from_check[0]', 'myappjarcontainer.jar'])
      assertDiracSucceeds(result, self)
      chdir_mock.assert_called_once_with('mytestarea51')
      exists_mock.assert_called_once_with('/current/dir/my/myappjarcontainer.jar')
      self.assertFalse(remove_mock.called)
      self.assertFalse(err_mock.called)

  def test_addfolder_to_ldlibrary(self):
    from ILCDIRAC.Core.Utilities.TARsoft import addFolderToLdLibraryPath
    with patch.dict(os.environ, {}, True):
      addFolderToLdLibraryPath('mytestfolder123')
      assertEqualsImproved(os.environ, {'LD_LIBRARY_PATH': 'mytestfolder123'}, self)

  def test_addfolder_to_ldlibrary_append(self):
    from ILCDIRAC.Core.Utilities.TARsoft import addFolderToLdLibraryPath
    with patch.dict(os.environ, {'LD_LIBRARY_PATH': 'prepath'}, True):
      addFolderToLdLibraryPath('other/TestFoldeR')
      assertEqualsImproved(os.environ, {'LD_LIBRARY_PATH': 'other/TestFoldeR:prepath'}, self)

  def test_install_package(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_OK(('APP_TAR', 'http://myTar.url', False, '1234MyMd5')))) as get_mock, \
        patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
        patch('%s.install' % MODULE_NAME, new=Mock(return_value=S_OK('res_from_install_test'))) as install_mock, \
        patch('%s.check' % MODULE_NAME, new=Mock(return_value=S_OK('res_from_check_test'))) as check_mock, \
        patch('%s.configure' % MODULE_NAME, new=Mock(return_value=S_OK())) as configure_mock, \
        patch('%s.clean' % MODULE_NAME, new=Mock(return_value=S_OK())) as clean_mock, \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('testAppLication', 'version1.01.'), 'configToTest', 'secret_area',
                                'mycurrenttestdir')
      assertDiracSucceeds(result, self)
      # change dir to current dir after every step
      assertMockCalls(chdir_mock, ['mycurrenttestdir'] * 5, self)
      get_mock.assert_called_once_with(('testAppLication', 'version1.01.'), 'configToTest', 'secret_area')
      install_mock.assert_called_once_with(('testAppLication', 'version1.01.'), 'APP_TAR',
                                            'http://myTar.url', False, '1234MyMd5', 'secret_area')
      check_mock.assert_called_once_with(('testAppLication', 'version1.01.'), 'secret_area',
                                          'res_from_install_test')
      configure_mock.assert_called_once_with(('testAppLication', 'version1.01.'), 'secret_area',
                                              'res_from_check_test')
      clean_mock.assert_called_once_with('secret_area', 'res_from_install_test')
      self.assertFalse(log_mock.called)

  def test_install_package_gettar_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_ERROR('tarball_err'))), \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('mytestappName', 'testv12'), 'configToTest', 'teAreast', 'mycurrenttestdir')
      assertDiracFailsWith(result, 'failed to install software', self)
      log_mock.assert_called_once_with(
          'Could not install software/dependency mytestappName testv12: tarball_err')

  def test_install_package_install_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_OK(('app_tar', 'tarballURL', False, 'md5sum')))), \
        patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.install' % MODULE_NAME, new=Mock(return_value=S_ERROR('install_test_err'))), \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('mytestappName', 'testv12'), 'configToTest', 'teAreast', 'mycurrenttestdir')
      assertDiracFailsWith(result, 'failed to install software', self)
      log_mock.assert_called_once_with(
          'Could not install software/dependency mytestappName testv12: install_test_err')

  def test_install_package_check_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_OK(('app_tar', 'tarballURL', False, 'md5sum')))), \
        patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.install' % MODULE_NAME, new=Mock(return_value=S_OK('res_from_install'))), \
        patch('%s.check' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_check_error_test'))), \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('mytestappName', 'testv12'), 'configToTest', 'teAreast', 'mycurrenttestdir')
      assertDiracFailsWith(result, 'failed to check integrity of software', self)
      log_mock.assert_called_with('Failed to check software/dependency mytestappName testv12')

  def test_install_package_configure_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_OK(('app_tar', 'tarballURL', False, 'md5sum')))), \
        patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.install' % MODULE_NAME, new=Mock(return_value=S_OK('res_from_install'))), \
        patch('%s.check' % MODULE_NAME, new=Mock(return_value=S_OK('mycheckresult_test'))), \
        patch('%s.configure' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_config_err_test'))), \
        patch('%s.clean' % MODULE_NAME, new=Mock(return_value=S_OK())), \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('mytestappName', 'testv12'), 'configToTest', 'teAreast', 'mycurrenttestdir')
      assertDiracFailsWith(result, 'failed to configure software', self)
      log_mock.assert_called_once_with('Failed to configure software/dependency mytestappName testv12')

  def test_install_package_clean_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import installPackage
    with patch('%s.getTarBallLocation' % MODULE_NAME, new=Mock(return_value=S_OK(('app_tar', 'tarballURL', False, 'md5sum')))), \
        patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.install' % MODULE_NAME, new=Mock(return_value=S_OK('res_from_install'))), \
        patch('%s.check' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.configure' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.clean' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_clean_err_test'))), \
            patch('%s.LOG.error' % MODULE_NAME) as log_mock:
      result = installPackage(('mytestappName', 'testv12'), 'configToTest', 'teAreast', 'mycurrenttestdir')
      assertDiracSucceeds(result, self)
      log_mock.assert_called_once_with(
          'Failed to clean useless tar balls, deal with it: mytestappName testv12')

  def test_install(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    extract_mock = Mock()
    extract_mock.getmembers.return_value = ['member0/file1.txt', 'member1']
    with patch('%s.os.chdir' % MODULE_NAME) as chdir_mock, \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(True))) as checklock_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])) as exists_mock, \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())) as createlock_mock, \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK())) as deleteold_mock, \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK())) as download_mock, \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=S_OK())) as clearlock_mock, \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_OK())) as md5check_mock, \
         patch('%s.tarfile.is_tarfile' % MODULE_NAME, new=Mock(return_value=False)) as tarcheck_mock, \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(side_effect=OSError(''))) as listdir_mock, \
         patch('%s.os.rename' % MODULE_NAME) as rename_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.os.unlink' % MODULE_NAME) as remove_mock:
      result = install(('appname', 'appvers'), 'myapptarbase.tar.gz', 'tarballURL', False, 'Mymd5sumTest',
                       'mytestArea')
      assertDiracSucceedsWith_equals(result, ['myapptarbase', 'myapptarbase.tar.gz'], self)
      clearlock_mock.assert_called_once_with('myapptarbase.lock')
      listdir_mock.assert_called_once_with('myapptarbase')
      self.assertFalse(rename_mock.called)
      self.assertFalse(remove_mock.called)
      self.assertFalse(deleteold_mock.called)
      createlock_mock.assert_called_once_with('myapptarbase.lock')
      download_mock.assert_called_once_with('tarballURL', 'myapptarbase.tar.gz', 'myapptarbase')
      tarcheck_mock.assert_called_once_with('myapptarbase.tar.gz')
      checklock_mock.assert_called_once_with('myapptarbase.lock')
      chdir_mock.assert_called_once_with('mytestArea')
      md5check_mock.assert_called_once_with('myapptarbase.tar.gz', 'Mymd5sumTest')
      assertMockCalls(exists_mock, ['myapptarbase', '/my/cwd/test/myapptarbase.tar.gz'], self)

  def test_install_checklockage_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_checklockage_failed'))):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', True, 'md5sum', 'area')
      assertDiracFailsWith(result, 'failed lock checks', self)

  def test_install_createlock_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_createlock_failed_err'))):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', True, 'md5sum', 'area')
      assertDiracFailsWith(result, 'test_createlock_failed_err', self)

  def test_install_already_installed(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracSucceedsWith_equals(result, ['app_tar', 'app_tar.tar.gz'], self)

  def test_install_delete_old_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_deleteold_file_error'))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock:
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', True, 'md5sum', 'area')
      assertDiracFailsWith(result, 'test_deleteold_file_error', self)
      clearlock_mock.assert_called_once_with('app_tar.lock')

  def test_install_download_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_download_file_error'))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock:
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'test_download_file_error', self)
      clearlock_mock.assert_called_once_with('app_tar.lock')

  def test_install_downloaded_file_doesnt_exist(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)) as exists_mock, \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'failed to download software', self)
      clearlock_mock.assert_called_once_with('app_tar.lock')
      assertMockCalls(exists_mock, ['app_tar', '/my/cwd/test/app_tar.tar.gz'], self)

  def test_install_md5check_cannot_remove_old_file(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
        patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
        patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
        patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
        patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
        patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_md5check_fails'))), \
        patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_ERROR('deleteold_test_error'))), \
        patch('%s.os.unlink' % MODULE_NAME, new=Mock(side_effect=OSError('my_os_err_test'))), \
            patch('%s.LOG.error' % MODULE_NAME, new=Mock(return_value=True)) as log_mock:
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'deleteold_test_error', self)
      log_mock.assert_called_with('Failed to clean tar ball, something bad is happening')
      clearlock_mock.assert_called_once_with('app_tar.lock')

  def test_install_md5check_download_file_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(side_effect=[S_OK(True), S_ERROR('download_file_test_err')])), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_md5check_fails'))), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.unlink' % MODULE_NAME):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'download_file_test_err', self)
      clearlock_mock.assert_called_once_with('app_tar.lock')

  def test_install_second_md5check_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_md5check_fails'))), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.unlink' % MODULE_NAME):
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'md5 check failed', self)
      clearlock_mock.assert_called_once_with('app_tar.lock')

  def test_install_tarfile_extract_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    from tarfile import TarError
    untar_me_mock = Mock()
    untar_me_mock.extractall.side_effect = TarError('custom_tar_test_err')
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.unlink' % MODULE_NAME), \
         patch('%s.tarfile.is_tarfile' % MODULE_NAME, new=Mock(return_value=True)) as istar_mock, \
         patch('%s.tarfile.open' % MODULE_NAME, new=Mock(return_value=untar_me_mock)) as tar_open_mock:
      result = install(('appname', 'appvers'), 'app_tar.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(
          result, 'Could not extract tar ball app_tar.tar.gz because of custom_tar_test_err, cannot continue ',
          self)
      clearlock_mock.assert_called_once_with('app_tar.lock')
      istar_mock.assert_called_once_with('app_tar.tar.gz')
      tar_open_mock.assert_called_once_with('app_tar.tar.gz')

  def test_install_tarfile_isslic_rename_fails(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    untar_member0 = Mock()
    untar_member0.name = 'member0/file1.txt'
    untar_me_mock = Mock()
    untar_me_mock.getmembers.return_value = [untar_member0, 'member1']
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.unlink' % MODULE_NAME), \
         patch('%s.tarfile.is_tarfile' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.tarfile.open' % MODULE_NAME, new=Mock(return_value=untar_me_mock)), \
         patch('%s.os.rename' % MODULE_NAME, new=Mock(side_effect=OSError('test_cannot_rename_err'))) as rename_mock:
      result = install(('appname', 'appvers'), 'slic.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'Could not rename slic directory', self)
      clearlock_mock.assert_called_once_with('slic.lock')
      rename_mock.assert_called_once_with('member0', 'slic')
      self.assertTrue(untar_me_mock.extractall.called)

  def test_install_tarfile_isslic_listdir_empty(self):
    from ILCDIRAC.Core.Utilities.TARsoft import install
    untar_member0 = Mock()
    untar_member0.name = 'member0/file1.txt'
    untar_me_mock = Mock()
    untar_me_mock.getmembers.return_value = [untar_member0, 'member1']
    with patch('%s.os.chdir' % MODULE_NAME), \
         patch('%s.checkLockAge' % MODULE_NAME, new=Mock(return_value=S_OK(False))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=[False, True])), \
         patch('%s.createLock' % MODULE_NAME, new=Mock(return_value=S_OK())), \
         patch('%s.downloadFile' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.clearLock' % MODULE_NAME, new=Mock(return_value=True)) as clearlock_mock, \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/cwd/test')), \
         patch('%s.tarMd5Check' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.deleteOld' % MODULE_NAME, new=Mock(return_value=S_OK(True))), \
         patch('%s.os.unlink' % MODULE_NAME), \
         patch('%s.tarfile.is_tarfile' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.tarfile.open' % MODULE_NAME, new=Mock(return_value=untar_me_mock)), \
         patch('%s.os.rename' % MODULE_NAME, new=Mock(return_value=True)) as rename_mock, \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=[])) as listdir_mock:
      result = install(('appname', 'appvers'), 'slic.tar.gz', 'tarballURL', False, 'md5sum', 'area')
      assertDiracFailsWith(result, 'folder ', self)
      clearlock_mock.assert_called_once_with('slic.lock')
      rename_mock.assert_called_once_with('member0', 'slic')
      self.assertTrue(untar_me_mock.extractall.called)
      listdir_mock.assert_called_once_with('slic')


MODULE_NAME = 'ILCDIRAC.Core.Utilities.TARsoft'



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
Whizard: First Generator application
"""

from __future__ import absolute_import
import types
import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.WhizardOptions import WhizardOptions, getDict
from ILCDIRAC.Core.Utilities.GeneratorModels import GeneratorModels

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class Whizard(LCApplication):
  """Runs whizard to generate a given event type.

  Usage:

  >>> wh = Whizard(dirac.getProcessList())
  >>> wh.setEvtType("ee_h_mumu")
  >>> wh.setEnergy(500)
  >>> wh.setNumberOfEvents(1000)
  >>> wh.setModel("sm")

  use :func:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard.Whizard.setExtraCLIArguments` to overwrite the content of the whizard.in
  in case you use something not standard (parameter scan for exmple)
  """

  def __init__(self, processlist=None, paramdict=None):

    self.parameterDict = {}
    self.model = 'sm'
    self.randomSeed = 0
    self.luminosity = 0
    self.jobIndex = ''
    self._optionsdictstr = ''
    self.fullParameterDict = {}
    self.generatorLevelCuts = {}
    self._genlevelcutsstr = ''
    self._leshouchesfiles = None
    self._generatormodels = GeneratorModels()
    self.eventType = ''
    self.globalEventType = ''
    self.useGridFiles = False
    self._allowedparams = ['PNAME1', 'PNAME2', 'POLAB1', 'POLAB2', 'USERB1', 'USERB2',
                           'ISRB1', 'ISRB2', 'EPAB1', 'EPAB2', 'RECOIL', 'INITIALS', 'USERSPECTRUM']
    self._wo = None
    self.parameters = []
    self._processlist = None
    if processlist:
      self._processlist = processlist
    super(Whizard, self).__init__(paramdict)
    # Those 4 need to come after default constructor
    self._modulename = 'WhizardAnalysis'
    self._moduledescription = 'Module to run WHIZARD'
    self.appname = 'whizard'
    self.datatype = 'gen'
    self._paramsToExclude.extend(['_optionsdictstr', '_genlevelcutsstr', '_leshouchesfiles', '_generatormodels',
                                    '_allowedparams', '_wo', '_processlist'])

  def getPDict(self):
    """Provide predefined parameter dictionary."""
    return getDict()

  def setEvtType(self, evttype):
    """Define process. If the process given is not found, when calling :func:`UserJob.append() <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.append>` a full list is printed.

    :param str evttype: Process to generate
    """
    self._checkArgs({'evttype': (str,)})
    if self.addedtojob:
      LOG.error("Cannot modify this attribute once application has been added to Job")
      return S_ERROR("Cannot modify")
    self.eventType = evttype

  def setGlobalEvtType(self, globalname):
    """When producing multiple process in one job, it is needed to define this for the output file name.

    It's mandatory to use the :any:`setFullParameterDict` method when using this.
    """
    self._checkArgs({'globalname': (str,)})
    self.globalEventType = globalname

  def setLuminosity(self, lumi):
    """ Optional: Define luminosity to generate

    :param float lumi: Luminosity to generate. Not available if cross section is not known a priori. Use with care.
    """
    self._checkArgs({'lumi': float})
    self.luminosity = lumi

  def setRandomSeed(self, randomSeed):
    """ Optional: Define random seed to use. Default is Job ID.

    :param int randomSeed: Seed to use during integration and generation.
    """
    self._checkArgs({'randomSeed': int})

    self.randomSeed = randomSeed

  def setParameterDict(self, paramdict):
    """Parameters for Whizard steering files.

    :param dict paramdict: Dictionary of parameters for the whizard templates. Most parameters are set on the fly.
    """
    self._checkArgs({'paramdict': dict})
    self.parameterDict = paramdict

  def setGeneratorLevelCuts(self, cutsdict):
    """Define generator level cuts (to be put in whizard.cut1)

    Refer to http://whizard.hepforge.org/manual_w1/manual005.html#toc12 for details about how to set cuts.

    >>> wh.setGeneratorLevelCuts({'e1e1_o':["cut M of  3 within 10 99999","cut E of  3 within  5 99999"]})

    :param dict cutsdict: Dictionary of cuts
    """
    self._checkArgs({'cutsdict': dict})
    self.generatorLevelCuts = cutsdict

  def setFullParameterDict(self, pdict):
    """Parameters for Whizard steering files, better than above as much more complete (cannot be more complete)

    >>> pdict = {}
    >>> pdict['process_input'] = {}
    >>> #processes below are not those of the templates, but those of the whizard.prc
    >>> pdict['process_input']['process_id']='h_n1n1'
    >>> pdict['process_input']['sqrts'] = 3000.
    >>> pdict['simulation_input'] = {}
    >>> pdict['simulation_input']['n_events'] = 100
    >>> pdict['beam_input_1'] = {}
    >>> pdict['beam_input_1']['polarization']='1.0 0.0'
    >>> pdict['beam_input_1']['USER_spectrum_mode'] = 11
    >>> pdict['beam_input_2'] = {}
    >>> pdict['beam_input_2']['polarization']='0.0 1.0'
    >>> pdict['beam_input_2']['USER_spectrum_mode'] = -11
    >>> wh.setFullParameterDict(pdict)

    The first key corresponds to the sections of the whizard.in, while the second corresponds to the possible parameters.
    All keys/values can be found in the WHIZARD documentation: http://whizard.hepforge.org/manual_w1/manual005.html#toc11

    :param dict pdict: Dictionnary of parameters
    """
    self._checkArgs({'pdict': dict})

    self.fullParameterDict = pdict
    # self._wo.changeAndReturn(dict)

  def setModel(self, model):
    """ Optional: Define Model

    :param str model: Model to use for generation. Predefined list available in the :mod:`GeneratorModels<ILCDIRAC.Core.Utilities.GeneratorModels.GeneratorModels>` class.
    """
    self._checkArgs({'model': (str,)})

    self.model = model

  def willCut(self):
    """You need this if you plan on cutting using :mod:`~ILCDIRAC.Interfaces.API.NewInterface.Applications.StdhepCut.StdhepCut`."""
    self.willBeCut = True

  def usingGridFiles(self):
    """Call this if you want to use the grid files that come with the Whizard version used.

    Beware: Depends on the energy and generator cuts, use it if you know what you are doing.
    """
    self.useGridFiles = True

  def setJobIndex(self, index):
    """ Optional: Define Job Index. Added in the file name between the event type and the extension.

    :param str index: Index to use for generation
    """
    self._checkArgs({'index': (str,)})

    self.jobIndex = index

  def dumpWhizardDotIn(self, fname='whizard.in'):
    """Dump the content of the whizard.in file requested for this application.

    :param str fname: filename in which whizard.in is written
    """
    if self.addedtojob:
      self._wo.toWhizardDotIn(fname)
    else:
      self._reportError("Can't dump the whizard.in as there can be further changes")

  def _checkConsistency(self, job=None):
    """Check the consistency, called from Application."""
    self._wo = WhizardOptions(self.model)

    if not self.fullParameterDict:
      if not self.energy:
        return S_ERROR('Energy not set')

      if not self.numberOfEvents:
        return S_ERROR('Number of events not set!')

      if not self.eventType:
        return S_ERROR("Process not defined")
    else:
      res = self._wo.checkFields(self.fullParameterDict)
      if not res['OK']:
        return res
      self._wo.changeAndReturn(self.fullParameterDict)
      res = self._wo.getValue("process_input/process_id")
      if not len(res['Value']):
        if self.eventType:
          if 'process_input' not in self.fullParameterDict:
            self.fullParameterDict['process_input'] = {}
          self.fullParameterDict['process_input']['process_id'] = self.eventType
        else:
          return S_ERROR("Event type not specified")
      self.eventType = res['Value']

      res = self._wo.getValue("process_input/sqrts")
      if isinstance(res['Value'], (int, int, float)):
        energy = res['Value']
      else:
        energy = eval(res['Value'])
      if not energy:
        if self.energy:
          if 'process_input' not in self.fullParameterDict:
            self.fullParameterDict['process_input'] = {}
          self.fullParameterDict['process_input']['sqrts'] = self.energy
          energy = self.energy
        else:
          return S_ERROR("Energy set to 0")
      self.energy = energy

      res = self._wo.getValue("simulation_input/n_events")
      if isinstance(res['Value'], (int, int, float)):
        numberOfEvents = res['Value']
      else:
        numberOfEvents = eval(res['Value'])
      if not numberOfEvents:
        if self.numberOfEvents:
          if 'simulation_input' not in self.fullParameterDict:
            self.fullParameterDict['simulation_input'] = {}
          self.fullParameterDict['simulation_input']['n_events'] = self.numberOfEvents
          numberOfEvents = self.numberOfEvents
        else:
          return S_ERROR("Number of events set to 0")
      self.numberOfEvents = numberOfEvents

    if not self._processlist:
      return S_ERROR("Process list was not given")

    if self.generatorLevelCuts:
      for process in self.generatorLevelCuts.keys():
        if process not in self.eventType.split():
          LOG.info("You want to cut on %s but that process is not to be generated" % process)
      for values in self.generatorLevelCuts.values():
        if not isinstance(values, list):
          return S_ERROR('Type of %s is not a list, cannot proceed' % values)
      self._genlevelcutsstr = str(self.generatorLevelCuts)

    if self.eventType:
      processes = self.eventType.split()
      if len(processes) > 1 and not self.globalEventType:
        return S_ERROR("Global name MUST be defined when producing multiple processes in one job")
      elif self.globalEventType:
        self.eventType = self.globalEventType
      for process in processes:
        if not self._processlist.existsProcess(process)['Value']:
          LOG.notice("Available processes are:")
          self._processlist.printProcesses()
          return S_ERROR('Process %s does not exists' % process)
        else:
          cspath = self._processlist.getCSPath(process)
          whiz_file = os.path.basename(cspath)
          version = whiz_file.replace(".tar.gz", "").replace(".tgz", "").replace("whizard", "")
          if self.version:
            if self.version != version:
              return S_ERROR("All processes to consider are not available in the same WHIZARD version")
          else:
            self.version = version
          LOG.info("Found the process %s in whizard %s" % (process, self.version))

    if not self.version:
      return S_ERROR('No version found')

    if self.model:
      if not self._generatormodels.hasModel(self.model)['OK']:
        return S_ERROR("Unknown model %s" % self.model)

    if self.outputFile:
      if self.outputFile.count("/"):
        return S_ERROR("The OutputFile name is a file name, not a path. Remove any / in there")

    if not self.outputFile and self._jobtype == 'User':
      self.outputFile = self.eventType
      if self.jobIndex:
        self.outputFile += "_" + self.jobIndex
      self.outputFile += "_gen.stdhep"

    if self._jobtype != 'User':
      if not self.willBeCut:
        self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                   "outputDataSE": '@{OutputSE}'})
      self.prodparameters['nbevts'] = self.numberOfEvents
      self.prodparameters['Process'] = self.eventType
      self.prodparameters['model'] = self.model
      self.prodparameters['Energy'] = self.energy
      self.prodparameters['whizardparams'] = self.fullParameterDict
      self.prodparameters['gencuts'] = self.generatorLevelCuts
      self.prodparameters['gridfiles'] = self.useGridFiles

    if not self.fullParameterDict and self.parameterDict:
      for key in self.parameterDict.keys():
        if key not in self._allowedparams:
          return S_ERROR("Unknown parameter %s" % key)

      self.setParameter('PNAME1', 'e1', "Assuming incoming beam 1 to be electrons")
      self.setParameter('PNAME2', 'E1', "Assuming incoming beam 2 to be positrons")
      self.setParameter('POLAB1', '0.0 0.0', "Assuming no polarization for beam 1")
      self.setParameter('POLAB2', '0.0 0.0', "Assuming no polarization for beam 2")
      self.setParameter('USERB1', 'T', "Will put beam spectrum to True for beam 1")
      self.setParameter('USERB2', 'T', "Will put beam spectrum to True for beam 2")
      self.setParameter('ISRB1', 'T', "Will put ISR to True for beam 1")
      self.setParameter('ISRB2', 'T', "Will put ISR to True for beam 2")

      self.setParameter('EPAB1', 'F', "Will put EPA to False for beam 1")
      self.setParameter('EPAB2', 'F', "Will put EPA to False for beam 2")

      self.setParameter('RECOIL', 'F', "Will set Beam_recoil to False")
      self.setParameter('INITIALS', 'F', "Will set keep_initials to False")
      self.setParameter('USERSPECTRUM', '11', "Will set USER_spectrum_on to +-11")

      self.parameters = ";".join(self.parameters)
    elif self.fullParameterDict:
      self._optionsdictstr = str(self.fullParameterDict)

    return S_OK()

  def _applicationModule(self):
    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("evttype", "", "string", "", "", False, False, "Process to generate"))
    md1.addParameter(Parameter("RandomSeed", 0, "int", "", "", False, False, "Random seed for the generator"))
    md1.addParameter(Parameter("Lumi", 0, "float", "", "", False, False, "Luminosity of beam"))
    md1.addParameter(Parameter("Model", "", "string", "", "", False, False, "Model for generation"))
    md1.addParameter(Parameter("SteeringFile", "", "string", "", "", False, False, "Steering file"))
    md1.addParameter(Parameter("steeringparameters", "", "string", "", "", False, False,
                               "Specific steering parameters"))
    md1.addParameter(Parameter("OptionsDictStr", "", "string", "", "", False, False,
                               "Options dict to create full whizard.in on the fly"))
    md1.addParameter(Parameter("GenLevelCutDictStr", "", "string", "", "", False, False,
                               "Generator level cuts to put in whizard.cut1"))
    md1.addParameter(Parameter("willCut", False, "bool", "", "", False, False, "Will cut after"))
    md1.addParameter(Parameter("useGridFiles", True, "bool", "", "", False, False, "Will use grid files"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):
    moduleinstance.setValue("evttype", self.eventType)
    moduleinstance.setValue("RandomSeed", self.randomSeed)
    moduleinstance.setValue("Lumi", self.luminosity)
    moduleinstance.setValue("Model", self.model)
    moduleinstance.setValue("SteeringFile", self.steeringFile)
    moduleinstance.setValue("steeringparameters", self.parameters)
    moduleinstance.setValue("OptionsDictStr", self._optionsdictstr)
    moduleinstance.setValue("GenLevelCutDictStr", self._genlevelcutsstr)
    moduleinstance.setValue("willCut", self.willBeCut)
    moduleinstance.setValue("useGridFiles", self.useGridFiles)
    moduleinstance.setValue("debug", self.debug)

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def setParameter(self, parameter, defaultValue, docString):
    if parameter not in self.parameterDict:
      LOG.info(docString)
      self.parameters.append("%s=%s" % (parameter, defaultValue))
    else:
      self.parameters.append("%s=%s" % (parameter, self.parameterDict[parameter]))



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Whizard analysis module.

Called by Job Agent.

:since: Sep 22, 2010

:author: S. Poss
"""

from __future__ import absolute_import
from DIRAC.Core.Utilities.Subprocess import shellCall
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder
from ILCDIRAC.Core.Utilities.ResolveDependencies import resolveDeps
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareWhizardFile
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareWhizardFileTemplate, getNewLDLibs
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from ILCDIRAC.Core.Utilities.ProcessList import ProcessList
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import getProdFilename
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Core.Utilities.GeneratorModels import GeneratorModels
from ILCDIRAC.Core.Utilities.WhizardOptions import WhizardOptions

from DIRAC import gLogger, S_OK, S_ERROR

import os
import shutil
import glob

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class WhizardAnalysis(ModuleBase):
  """Specific Module to run a Whizard job."""

  def __init__(self):
    super(WhizardAnalysis, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.debug = True
    self.SteeringFile = ''
    self.OutputFile = ''
    self.NumberOfEvents = 1
    self.Lumi = 0
    self.applicationName = 'whizard'
    self.evttype = ""
    self.RandomSeed = 0
    self.getProcessInFile = False
    self.datMan = DataManager()
    self.processlist = None
    self.parameters = {}
    self.susymodel = 0
    self.Model = ''
    self.genmodel = GeneratorModels()
    self.eventstring = ['! ', 'Fatal error:', 'PYSTOP', 'No matrix element available',
                        'Floating point exception', 'Event generation finished.', " n_events", "luminosity",
                        "  sum            "]
    self.excludeAllButEventString = False
    self.steeringparameters = ''
    self.options = None
    self.optionsdict = {}
    self.OptionsDictStr = ''
    self.GenLevelCutDictStr = ''
    self.genlevelcuts = {}
    self.willCut = False
    self.useGridFiles = False

  def obtainProcessList(self):
    """Internal function.

    Get the process list from storage if whizard.in was not provided

    :return: S_OK(), S_ERROR()
    """

    res = self.ops.getValue("/ProcessList/Location", "")
    if not res:
      return S_ERROR("No process list found")
    processlistloc = res
    if not os.path.exists(os.path.basename(processlistloc)):
      res = self.datMan.getFile(processlistloc)
      if not res['OK']:
        LOG.error('Could not get processlist: %s' % res['Message'])
        return res
    self.processlist = ProcessList(os.path.basename(processlistloc))
    return S_OK()

  def applicationSpecificInputs(self):
    """Resolve module input.

    :return: S_OK()
    """
    self.parameters['ENERGY'] = self.energy

    if not self.RandomSeed and self.jobID:
      self.RandomSeed = self.jobID
    if 'IS_PROD' in self.workflow_commons or 'IS_DBD_GEN_PROD' in self.workflow_commons:
      self.RandomSeed = int(
          str(
              int(
                  self.workflow_commons["PRODUCTION_ID"]))
          + str(
              int(
                  self.workflow_commons["JOB_ID"])))

    self.parameters['SEED'] = self.RandomSeed
    self.parameters['NBEVTS'] = self.NumberOfEvents
    self.parameters['LUMI'] = self.Lumi

    # EVER USED???
    if 'SusyModel' in self.step_commons:
      self.susymodel = self.step_commons['SusyModel']

    self.SteeringFile = os.path.basename(self.step_commons.get("InputFile", self.SteeringFile))
    if self.SteeringFile == "whizard.in":
      os.rename(self.SteeringFile, "whizardnew.in")
      self.SteeringFile = "whizardnew.in"

    self.parameters['PROCESS'] = self.evttype

    listofparams = self.steeringparameters.split(";")
    for param in listofparams:
      if param.count("="):
        self.parameters[param.split("=")[0]] = param.split("=")[1]

    if self.OptionsDictStr:
      LOG.info("Will use whizard.in definition from WhizardOptions.")
      try:
        self.optionsdict = eval(self.OptionsDictStr)
        if 'integration_input' not in self.optionsdict:
          self.optionsdict['integration_input'] = {}
        if 'seed' not in self.optionsdict['integration_input']:
          self.optionsdict['integration_input']['seed'] = int(self.RandomSeed)
        if 'process_input' in self.optionsdict:
          if 'sqrts' in self.optionsdict['process_input']:
            self.energy = self.optionsdict['process_input']['sqrts']
      except BaseException:
        return S_ERROR("Could not convert string to dictionary for optionsdict")

    if self.GenLevelCutDictStr:
      LOG.info("Found generator level cuts")
      try:
        self.genlevelcuts = eval(self.GenLevelCutDictStr)
      except BaseException:
        return S_ERROR("Could not convert the generator level cuts back to dictionary")

    if not len(self.SteeringFile) and not self.optionsdict:
      self.getProcessInFile = True

    if "IS_PROD" in self.workflow_commons:
      if self.workflow_commons["IS_PROD"] and not self.willCut:
        # self.OutputFile = getProdFilename(self.OutputFile,int(self.workflow_commons["PRODUCTION_ID"]),
        #                                  int(self.workflow_commons["JOB_ID"]))
        if 'ProductionOutputData' in self.workflow_commons:
          outputlist = self.workflow_commons['ProductionOutputData'].split(";")
          for obj in outputlist:
            if obj.lower().count("_gen_"):
              self.OutputFile = os.path.basename(obj)
              break
        else:
          # This is because most likely there is stdhepcut running after
          self.OutputFile = "willcut.stdhep"

          # getProdFilename(self.OutputFile,int(self.workflow_commons["PRODUCTION_ID"]),
          #                                  int(self.workflow_commons["JOB_ID"]))

    if "IS_DBD_GEN_PROD" in self.workflow_commons and self.workflow_commons["IS_DBD_GEN_PROD"]:
      # self.OutputFile = getProdFilename(self.OutputFile,int(self.workflow_commons["PRODUCTION_ID"]),
      #                                  int(self.workflow_commons["JOB_ID"]))
      if 'ProductionOutputData' in self.workflow_commons:
        outputlist = self.workflow_commons['ProductionOutputData'].split(";")
        for obj in outputlist:
          self.OutputFile = os.path.basename(obj)
          break
      else:
        self.OutputFile = getProdFilename(self.OutputFile, int(self.workflow_commons["PRODUCTION_ID"]),
                                          int(self.workflow_commons["JOB_ID"]))
    return S_OK()

  def runIt(self):
    """Called by Agent.

    Executes the following
      - resolve input variables
      - resolve installation location
      - resolve dependencies location (beam_spectra)
      - get processlist if needed
      - define output file name
      - prepare whizard.in
      - make magic

    :return: S_OK(), S_ERROR()
    """
    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result["Message"])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('Whizard should not proceed as previous step did not end properly')

    # if self.debug:
    #  self.excludeAllButEventString = False

    res = getSoftwareFolder(self.platform, self.applicationName, self.applicationVersion)
    if not res['OK']:
      LOG.error("Failed getting software folder", res['Message'])
      self.setApplicationStatus('Failed finding software')
      return res
    mySoftDir = res['Value']

    # Remove libc
    removeLibc(mySoftDir + "/lib")

    # Need to fetch the new LD_LIBRARY_PATH
    new_ld_lib_path = getNewLDLibs(self.platform, self.applicationName, self.applicationVersion)
    # Don't forget to prepend the application's libs
    new_ld_lib_path = mySoftDir + "/lib:" + new_ld_lib_path
    # Resolve dependencies (look for beam_spectra)
    deps = resolveDeps(self.platform, self.applicationName, self.applicationVersion)
    path_to_beam_spectra = ""
    path_to_gridfiles = ""
    for dep in deps:
      res = getSoftwareFolder(self.platform, dep["app"], dep['version'])
      if not res['OK']:
        LOG.error("Failed getting software folder", res['Message'])
        self.setApplicationStatus('Failed finding software')
        return res
      depfolder = res['Value']
      if dep["app"] == "beam_spectra":
        path_to_beam_spectra = depfolder
      elif dep["app"] == "gridfiles":
        path_to_gridfiles = depfolder

    # Env variables needed to run whizard: avoids hard coded locations
    os.environ['LUMI_LINKER'] = path_to_beam_spectra + "/lumi_linker_000"
    os.environ['PHOTONS_B1'] = path_to_beam_spectra + "/photons_beam1_linker_000"
    os.environ['PHOTONS_B2'] = path_to_beam_spectra + "/photons_beam2_linker_000"
    os.environ['EBEAM'] = path_to_beam_spectra + "/ebeam_in_linker_000"
    os.environ['PBEAM'] = path_to_beam_spectra + "/pbeam_in_linker_000"

    os.environ['LUMI_EE_LINKER'] = path_to_beam_spectra + "/lumi_ee_linker_000"
    os.environ['LUMI_EG_LINKER'] = path_to_beam_spectra + "/lumi_eg_linker_000"
    os.environ['LUMI_GE_LINKER'] = path_to_beam_spectra + "/lumi_ge_linker_000"
    os.environ['LUMI_GG_LINKER'] = path_to_beam_spectra + "/lumi_gg_linker_000"

    list_of_gridfiles = []
    if path_to_gridfiles and self.useGridFiles:
      tmp_list_of_gridfiles = [os.path.join(path_to_gridfiles, item) for item in os.listdir(path_to_gridfiles)]
      gridfilesfound = False
      for path in tmp_list_of_gridfiles:
        if os.path.isdir(path) and path.count(str(self.energy)):
          # Here look for a sub directory for the energy related grid files
          list_of_gridfiles = [os.path.join(path, item) for item in os.listdir(path)]
          gridfilesfound = True
          LOG.info('Found grid files specific for energy %s' % self.energy)
          break
      if not gridfilesfound:
        LOG.info("Will use generic grid files found, hope the energy is set right")
        list_of_gridfiles = [
            item for item in glob.glob(
                os.path.join(
                    path_to_gridfiles,
                    "*.grb"))
            + glob.glob(
                os.path.join(
                    path_to_gridfiles,
                    "*.grc"))]

    template = False
    if self.SteeringFile.count("template"):
      template = True
    # Get from process file the proper whizard.in file
    if self.getProcessInFile:
      whizardin = ""
      res = self.obtainProcessList()
      if not res['OK']:
        LOG.error("Could not obtain process list")
        self.setApplicationStatus('Failed getting processlist')
        return res
      whizardin = self.processlist.getInFile(self.evttype)
      if not whizardin:
        LOG.error("Whizard input file was not found in process list, cannot proceed")
        self.setApplicationStatus('Whizard input file was not found')
        return S_ERROR("Error while resolving whizard input file")
      if whizardin.count("template"):
        template = True
      try:
        shutil.copy("%s/%s" % (mySoftDir, whizardin), "./whizardnew.in")
        self.SteeringFile = "whizardnew.in"
      except EnvironmentError:
        LOG.error("Could not copy %s from %s" % (whizardin, mySoftDir))
        self.setApplicationStatus('Failed getting whizard.in file')
        return S_ERROR("Failed to obtain %s" % whizardin)

    # Check existence of Les Houches input file
    leshouchesfiles = ''
    if not os.path.exists("LesHouches.msugra_1.in"):
      if self.susymodel:
        if self.susymodel == 1:
          if os.path.exists("%s/LesHouches_slsqhh.msugra_1.in" % (mySoftDir)):
            leshouchesfiles = "%s/LesHouches_slsqhh.msugra_1.in" % (mySoftDir)
        if self.susymodel == 2:
          if os.path.exists("%s/LesHouches_chne.msugra_1.in" % (mySoftDir)):
            leshouchesfiles = "%s/LesHouches_chne.msugra_1.in" % (mySoftDir)
      if self.Model:
        if self.genmodel.hasModel(self.Model)['OK']:
          if self.genmodel.getFile(self.Model)['OK']:
            if os.path.exists("%s/%s" % (mySoftDir, self.genmodel.getFile(self.Model)['Value'])):
              leshouchesfiles = "%s/%s" % (mySoftDir, self.genmodel.getFile(self.Model)['Value'])
            else:
              LOG.error("Request LesHouches file is missing, cannot proceed")
              self.setApplicationStatus("LesHouches file missing")
              return S_ERROR("The LesHouches file was not found. Probably you are using a wrong version of whizard.")
          else:
            LOG.warn("No file found attached to model %s" % self.Model)
        else:
          LOG.error("Model undefined:", self.Model)
          self.setApplicationStatus("Model undefined")
          return S_ERROR("No Model %s defined" % self.Model)
    else:
      leshouchesfiles = "LesHouches.msugra_1.in"

    outputfilename = self.evttype

    if self.optionsdict:
      LOG.info("Using: %s" % self.optionsdict)
      self.options = WhizardOptions(self.Model)
      res = self.options.changeAndReturn(self.optionsdict)
      if not res['OK']:
        return res
      res = self.options.toWhizardDotIn("whizard.in")
    elif not template:
      res = prepareWhizardFile(self.SteeringFile, outputfilename, self.energy,
                               self.RandomSeed, self.NumberOfEvents, self.Lumi,
                               "whizard.in")
    else:
      res = prepareWhizardFileTemplate(self.SteeringFile, outputfilename,
                                       self.parameters, "whizard.in")
    if not res['OK']:
      LOG.error('Something went wrong with input file generation')
      self.setApplicationStatus('Whizard: something went wrong with input file generation')
      return S_ERROR('Something went wrong with whizard.in file generation')
    foundproceesinwhizardin = res['Value']

    scriptName = 'Whizard_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = open(scriptName, 'w')
    script.write('#!/bin/sh \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    script.write('declare -x PATH=%s:$PATH\n' % mySoftDir)
    script.write('declare -x LD_LIBRARY_PATH=%s\n' % new_ld_lib_path)
    script.write('env | sort >> localEnv.log\n')
    script.write('echo =============================\n')
    script.write('echo Printing content of whizard.in \n')
    script.write('cat whizard.in\n')
    script.write('echo =============================\n')
    script.write('cp  %s/whizard.mdl ./\n' % mySoftDir)
    if leshouchesfiles:
      if not leshouchesfiles == 'LesHouches.msugra_1.in':
        script.write('cp %s ./LesHouches.msugra_1.in\n' % (leshouchesfiles))
      script.write('ln -s LesHouches.msugra_1.in fort.71\n')
    if len(list_of_gridfiles):
      for gridfile in list_of_gridfiles:
        script.write('cp %s ./\n' % (gridfile))
    script.write('cp %s/whizard.prc ./\n' % mySoftDir)
    if self.genlevelcuts:
      res = self.makeWhizardDotCut1()
      if not res['OK']:
        script.close()
        LOG.error("Could not create the cut1 file")
        return S_ERROR("Could not create the cut1 file")
    script.write('echo =============================\n')
    script.write('echo Printing content of whizard.prc \n')
    script.write('cat whizard.prc\n')
    script.write('echo =============================\n')
    extracmd = ""
    if not self.debug:
      extracmd = "2>/dev/null"

    comm = ""
    if foundproceesinwhizardin:
      comm = 'whizard --simulation_input \'write_events_file = \"%s\"\'' % (outputfilename)
    else:
      comm = 'whizard --process_input \'process_id =\"%s\"\' --simulation_input \'write_events_file = \"%s\"\' ' % (
          self.evttype, outputfilename)
    comm = "%s %s %s\n" % (comm, self.extraCLIarguments, extracmd)
    LOG.info("Will run %s" % comm)
    script.write(comm)
    script.write('declare -x appstatus=$?\n')
    script.write('exit $appstatus\n')

    script.close()
    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)
    os.chmod(scriptName, 0o755)
    comm = 'sh -c "./%s"' % (scriptName)
    self.setApplicationStatus('Whizard %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=209715200)
    #self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
    if not self.result['OK']:
      LOG.error("Failed with error %s" % self.result['Message'])
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
      if not self.ignoreapperrors:
        return S_ERROR('%s did not produce the expected log' % (self.applicationName))
    lumi = ''
    message = ""
    success = False
    # Analyse log file
    with open(self.applicationLog) as logfile:
      for line in logfile:
        if line.count('! Event sample corresponds to luminosity'):
          elems = line.split()
          lumi = elems[-1]
        if line.count("*** Fatal error:"):
          status = 1
          message = line
          break
        elif line.count("PYSTOP"):
          status = 1
          message = line
          break
        elif line.count("No matrix element available"):
          status = 1
          message = line
          break
        elif line.count("Floating point exception"):
          status = 1
          message = line
          break
        elif line.count("Event generation finished."):
          success = True
        else:
          status = 0
    if success:
      status = 0
    else:
      status = 1
    LOG.info('The sample generated has an equivalent luminosity of %s' % lumi)
    if lumi:
      self.workflow_commons['Luminosity'] = float(lumi)
    else:
      status = 1

    # Now care for the cross sections
    info = {}
    res = self.options.getAsDict()
    if os.path.exists("whizard.out") and res['OK']:
      full_opts_dict = res['Value']
      processes = full_opts_dict['process_input']['process_id'].split()
      info = {}
      info['xsection'] = {}
      processes.append('sum')
      with open("whizard.out", "r") as inf:
        for line in inf:
          line = line.rstrip()
          for process in processes:
            if not process:
              continue
            if line.count("   %s            " % process):
              info['xsection'][process] = {}
              line = line.lstrip()
              crosssection = line.split()[1]
              err_crosssection = line.split()[2]
              frac = line.split()[4]
              info['xsection'][process]['xsection'] = float(crosssection)
              info['xsection'][process]['err_xsection'] = float(err_crosssection)
              info['xsection'][process]['fraction'] = float(frac)

    if info:
      if 'Info' not in self.workflow_commons:
        self.workflow_commons['Info'] = info
      else:
        self.workflow_commons['Info'].update(info)

    LOG.info("Status after the application execution is %s" % str(status))

    messageout = 'Whizard %s Successful' % (self.applicationVersion)
    failed = False
    if status != 0:
      LOG.error("Whizard execution completed with errors:")
      failed = True
    else:
      LOG.info("Whizard execution completed successfully")
      # Deal with output file
      if len(self.OutputFile):
        if os.path.exists(outputfilename + ".001.stdhep"):
          LOG.notice("Looking for output files")
          ofnames = glob.glob(outputfilename + '*.stdhep')
          if len(ofnames) > 1:
            basename = self.OutputFile.split(".stdhep")[0]
            i = 0
            for of in ofnames:
              i += 1
              name = basename + "_" + str(i) + ".stdhep"
              os.rename(of, name)
          else:
            os.rename(outputfilename + ".001.stdhep", self.OutputFile)
        else:
          LOG.error("Whizard execution did not produce a stdhep file")
          self.setApplicationStatus('Whizard %s Failed to produce STDHEP file' % (self.applicationVersion))
          messageout = 'Whizard Failed to produce STDHEP file'
          if not self.ignoreapperrors:
            return S_ERROR(messageout)

    if failed is True:
      LOG.error("==================================\n StdError:\n")
      LOG.error(message)
      self.setApplicationStatus('%s Exited With Status %s' % (self.applicationName, status))
      LOG.error('Whizard Exited With Status %s' % (status))
      messageout = 'Whizard Exited With Status %s' % (status)
      if not self.ignoreapperrors:
        return S_ERROR(messageout)
    else:
      self.setApplicationStatus(messageout)
    return S_OK({"OutputFile": self.OutputFile})

  def makeWhizardDotCut1(self):
    """When users need whizard cuts, this is called to prepare the file.

    :return: S_OK()
    """
    cutf = open("whizard.cut1", "w")
    for key, values in self.genlevelcuts.items():
      cutf.write("process %s\n" % key)
      for val in values:
        cutf.write("  %s\n" % val)
    cutf.close()
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Created on Feb 8, 2012 Modified on Aug 17, 2012.

:author: Stephane Poss, Christian Grefe
"""

from __future__ import print_function
from __future__ import absolute_import
from DIRAC.Core.Base import Script
import sys
from six.moves import input

# default values for the parameters
prodID = None
process = None
dataType = 'GEN'
machine = 'ilc'
detectorType = 'sid'
detectorModel = 'sidloi3'
polarisation = None  # can only be 'p80m20' or 'm80p20' in the DBD context
nEvtsPerJob = 100
energy = 1000
nameAppendix = ''
groupPrefix = 'SiD_DBD'
slicVersion = 'v3r0p3'  # has to exist in the DIRAC configuration
stdhepSplitVersion = 'V2'  # has to exist in the DIRAC configuration
slicMacro = 'defaultIlcCrossingAngle.mac'  # has to exist in the default steering files tarball available in DIRAC
outputSRM = 'RAL-SRM'
replicaSRMs = ["SLAC-SRM", "FNAL-SRM"]


# definitions for command line parsing
Script.registerSwitch('D:', 'detector=', 'detector model used for the simulation (default: %s)' % (detectorModel))
Script.registerSwitch('e:', 'energy=', 'energy used to generate input files (default: %s)' % (energy))
Script.registerSwitch('i:', 'prodid=', 'production id of the input files')
Script.registerSwitch('m:', 'machine=', 'machine used to generate input files (default: %s)' % (machine))
Script.registerSwitch('n:', 'nevents=', 'maximum number of events per job (default: %s)' % (nEvtsPerJob))
Script.registerSwitch('N:', 'name=', 'appended to the default production name')
Script.registerSwitch('p:', 'process=', 'event type of the input files')
Script.registerSwitch('P:', 'polarisation=', 'polarisation used to generate input files (default: %s)' % (polarisation))
Script.registerSwitch('S:', 'storage=', 'storage element used to save the output files (default: %s)' % (outputSRM))
Script.registerSwitch('V:', 'slicversion=', 'slic version used in the simulation (default: %s)' % (slicVersion))

Script.setUsageMessage(sys.argv[0] + '-i <prodID> -p <process> -P <polarisation>')

Script.parseCommandLine()

# process the parsed switches
switches = Script.getUnprocessedSwitches()
for switch in switches:
  opt = switch[0]
  arg = switch[1]
  if opt in ('D', 'detector'):
    detectorModel = arg
  if opt in ('e', 'energy'):
    energy = arg
  if opt in ('i', 'prodid'):
    prodID = arg
  if opt in ('m', 'machine'):
    machine = arg
  if opt in ('n', 'nevents'):
    nEvtsPerJob = arg
  if opt in ('N', 'name'):
    nameAppendix = arg
  if opt in ('p', 'process'):
    process = arg
  if opt in ('P', 'polarisation'):
    polarisation = arg
  if opt in ('S', 'storage'):
    storage = arg
  if opt in ('V', 'slicversion'):
    slicversion = arg

from ILCDIRAC.Interfaces.API.NewInterface.SIDProductionJob import SIDProductionJob
from ILCDIRAC.Interfaces.API.NewInterface.Applications import SLIC, StdHepSplit

# As it's a full chain, we start at generation
# so we need to define the process and the energy
# The rest will be set later. We could also set the process
# and the energy directly in the whizard def, but for clarity
# it's better to do it before, that way we know the very
# essential

if not prodID or not process or not polarisation:
  print('\tNeed to define at least production ID, process name and polarisation name.')
  print('\tUse -h for help.')
  sys.exit(2)

# define a meta data dictionary for lookup in the file catalog
meta = {}
meta['ProdID'] = prodID
meta['EvtType'] = process
#meta['Energy'] = energy
meta['Datatype'] = dataType
meta['Machine'] = machine
meta['DetectorType'] = detectorType
meta['Polarisation'] = polarisation

# give feedback on the meta data
print('Creating production using meta data query:')
for key in meta:
  print('  %s:\t%s' % (key, meta[key]))

# confirm the meta data query
res = ''
while res not in ['y', 'Y']:
  res = input('\nContinue? (y/n)\t')
  if res in ['n', 'N']:
    sys.exit(0)

# Do Split
activesplitstdhep = True

# Do Sim
sid_sim = True

# Do Replication
replicateFiles = True

# Split
stdhepsplit = StdHepSplit()
stdhepsplit.setVersion(stdhepSplitVersion)
stdhepsplit.setNumberOfEventsPerFile(nEvtsPerJob)

# Simulation
slic = SLIC()
slic.setVersion(slicVersion)
slic.setSteeringFile(slicMacro)
slic.setDetectorModel(detectorModel)
slic.setNumberOfEvents(nEvtsPerJob)

############################################
# What is below WILL NEVER NEED TO BE TOUCHED
# (I'm not kidding, if you touch and break, not my problem)
#
# Define production step splitting the stdhep files
if activesplitstdhep and meta:
  pstdhepsplit = SIDProductionJob()
  pstdhepsplit.setLogLevel('verbose')
  pstdhepsplit.setProdType('Split')
  res = pstdhepsplit.setInputDataQuery(meta)
  if not res['OK']:
    print(res['Message'])
    exit(1)
  pstdhepsplit.setOutputSE(outputSRM)
  wname = '%s_%s_%s_split' % (process, energy, polarisation)
  if nameAppendix:
    wname += '_%s' % (nameAppendix)
  pstdhepsplit.setWorkflowName(wname)
  pstdhepsplit.setProdGroup('%s_%s_%s' % (groupPrefix, process, energy))

  # Add the application
  res = pstdhepsplit.append(stdhepsplit)
  if not res['OK']:
    print(res['Message'])
    exit(1)
  pstdhepsplit.addFinalization(True, True, True, True)
  descrp = 'Splitting stdhep files'
  if nameAppendix:
    descrp += ', %s' % (nameAppendix)
  pstdhepsplit.setDescription(descrp)

  res = pstdhepsplit.createProduction()
  if not res['OK']:
    print(res['Message'])
  res = pstdhepsplit.finalizeProd()
  if not res['OK']:
    print(res['Message'])
    exit(1)
  # get the metadata for this production to define input for the next step
  meta = pstdhepsplit.getMetadata()


# Define a production step for the simulation
if sid_sim and meta:
  psl = SIDProductionJob()
  psl.setLogLevel('verbose')
  psl.setProdType('MCSimulation')
  res = psl.setInputDataQuery(meta)
  if not res['OK']:
    print(res['Message'])
    exit(1)
  psl.setOutputSE(outputSRM)
  wname = '%s_%s_%s_%s_sim' % (process, energy, polarisation, detectorModel)
  if nameAppendix:
    wname += '_%s' % (nameAppendix)
  psl.setWorkflowName(wname)
  psl.setProdGroup('%s_%s_%s' % (groupPrefix, process, energy))
  # Add the application
  res = psl.append(slic)
  if not res['OK']:
    print(res['Message'])
    exit(1)
  psl.addFinalization(True, True, True, True)
  descrp = '%s model' % (detectorModel)
  if nameAppendix:
    descrp += ', %s' % (nameAppendix)
  psl.setDescription(descrp)

  res = psl.createProduction()
  if not res['OK']:
    print(res['Message'])
  res = psl.finalizeProd()
  if not res['OK']:
    print(res['Message'])
    exit(1)
  # get the metadata for this production to define input for the next step
  meta = psl.getMetadata()


from DIRAC.TransformationSystem.Client.Transformation import Transformation
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

# Define transformation steps for the replication of the output data
if replicateFiles and meta:
  Trans = Transformation()
  Trans.setTransformationName('replicate_%s_%s_%s_%s' % (process, energy, polarisation, meta['Datatype']))
  description = 'Replicate %s %s %s %s to' % (process, energy, polarisation, meta['Datatype'])
  for replicaSRM in replicaSRMs:
    description += ' %s,' % (replicaSRM)
  description.rstrip(',')
  Trans.setDescription(description)
  Trans.setLongDescription(description)
  Trans.setType('Replication')
  Trans.setPlugin('Broadcast')
  Trans.setSourceSE(outputSRM)
  Trans.setTargetSE(replicaSRMs)

  res = Trans.addTransformation()
  if not res['OK']:
    print(res)
    sys.exit(0)
  print(res)
  Trans.setStatus('Active')
  Trans.setAgentType('Automatic')
  currtrans = Trans.getTransformationID()['Value']
  client = TransformationClient()
  res = client.createTransformationInputDataQuery(currtrans, meta)
  print(res['OK'])



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Mixin for userjob splitting options."""

from __future__ import absolute_import
import types

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.List import breakListIntoChunks

from ILCDIRAC.Core.Utilities.Utilities import toInt
from six.moves import range

LOG = gLogger.getSubLogger(__name__)


class SplitMixin(object):
  """Mixin class for user job splitting functions."""

  def _initialize(self):
    """Initialize splitting members."""
    self._data = []
    self._splittingOption = None
    self._switch = {'byEvents': self._splitByEvents,
                    'byData': self._splitByData,
                    'bySkip': self._splitBySkip,
                    }

    self._numberOfJobs = None
    self._totalNumberOfEvents = None
    self._eventsPerJob = None
    self._eventsPerFile = None
    self._numberOfFilesPerJob = 1
    self._startJobIndex = 0
    self._jobIndexList = None

  def _checkSplitConsistency(self):
    """Check if splitting options are consistent.

    :return: The success or the failure of the consistency checking
    :rtype: bool
    """
    LOG.notice("Checking job consistency")

    if self._splittingOption and self._splittingOption not in self._switch:
      splitOptions = ",".join(list(self._switch.keys()))
      return self._reportError("Consistency check failed: Bad split value: possible values are %s" % splitOptions)

    # All applications should have the same number of events
    # We can get this number from the first application for example
    sameNumberOfEvents = next(iter(self.applicationlist)).numberOfEvents

    if not all(app.numberOfEvents == sameNumberOfEvents for app in self.applicationlist):
      return self._reportError("Job: Applications should all have the same number of events")

    if sameNumberOfEvents == -1 and not self._data:
      return self._reportError("Job: Number of events is -1 without input data. Was that intentional?")

    LOG.notice("Consistency check successful")
    return S_OK()

  def setSplitEvents(self, eventsPerJob=None, numberOfJobs=None, totalNumberOfEvents=None):
    """Set up to run a given number of jobs producing a certain number of events.

    Example usage:

    >>> job = UserJob()
    >>> job.setSplitEvents(numberOfJobs=42, totalNumberOfEvents=126)

    Exactly two of the parmeters should be set

    :param int eventsPerJob: The events processed by a single job
    :param int numberOfJobs: The number of jobs
    :param int totalNumberOfEvents: The total number of events processed by all jobs

    .. note::

       This functionality is meant to simulate particle guns or for running MC
       generators, where the only parameter that change are the random seed and
       the output filename.
    """
    self._totalNumberOfEvents = totalNumberOfEvents
    self._eventsPerJob = eventsPerJob
    self._numberOfJobs = numberOfJobs

    self._addParameter(self.workflow, 'NbOfEvts', 'JDL', -1, 'Number of Events')

    self._splittingOption = "byEvents"

  def setSplitInputData(self, lfns, numberOfFilesPerJob=1):
    """Set up to run over a list of input files.

    Example usage:

    >>> job = UserJob()
    >>> job.setSplitInputData(listOfLFNs, numberOfFilesPerJob=10)

    :param lfns: Logical File Names
    :type lfns: list(str)
    :param int numberOfFilesPerJob: The number of input data processed by a single job
    """
    self._data = lfns if isinstance(lfns, list) else [lfns]
    self._numberOfFilesPerJob = numberOfFilesPerJob

    self._splittingOption = 'byData'

  def setSplitFilesAcrossJobs(self, lfns, eventsPerFile, eventsPerJob):
    """Split a list of input files into many jobs per file.

    :param lfns: Logical File Names
    :type lfns: list(str)
    :param int eventsPerFile: number of evenst in each file
    :param int eventsPerJob: number of events in each job. There will be eventsPerFile/eventsPerJob jobs for each file
    """
    self._data = lfns if isinstance(lfns, list) else [lfns]
    self._eventsPerFile = eventsPerFile
    self._eventsPerJob = eventsPerJob
    self._splittingOption = 'bySkip'
    self._addParameter(self.workflow, 'NbOfEvts', 'JDL', -1, 'Number of Events')

  def setSplitDoNotAlterOutputFilename(self, value=True):
    """Disable the changing the output filename for jobs with splitting.

    If this option is set the output data lfns will _not_ include the JobIndex

    :param bool value: if *True* disable the changing of the output data
        filenames. If *False* the JobIndex will be added at the end of
        OutputData LFNs before the extension. Or replace '%n' with the jobIndex
        in the fileName. See :func:`~ILCDIRAC.Core.Utilities.Splitting.addJobIndexToFilename`
    """
    self._addParameter(self.workflow, 'DoNotAlterOutputData', 'JDL', value, 'Do Not Change Output Data')

  def _split(self):
    """Check the consistency of the job and call the right split method.

    :return: The success or the failure of the consistency checking, :func`~DIRAC.Core.Utilities.ReturnValues.S_OK`, :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    LOG.notice("Job splitting...")

    self._eventsPerJob = toInt(self._eventsPerJob, cond=lambda x: x >= 0)
    self._numberOfJobs = toInt(self._numberOfJobs, cond=lambda x: x >= 0)

    if self._numberOfJobs is False or self._eventsPerJob is False:
      return self._reportError("Splitting: Invalid values for splitting: %s, %s" %
                               (self._eventsPerJob, self._numberOfJobs))

    resCheck = self._checkSplitConsistency()
    if not resCheck['OK']:
      return resCheck

    sequences = self._switch.get(self._splittingOption, lambda: [])()
    sequenceLength = 0
    for sequenceType, sequenceList, addToWorkflow in sequences:
      LOG.debug("Adding sequence %s %s" % (sequenceType, addToWorkflow))
      self.setParameterSequence(sequenceType, sequenceList, addToWorkflow)
      sequenceLength = len(sequenceList)

    if sequences:
      LOG.debug("Adding jobIndexList")
      if not self._jobIndexList:
        self._jobIndexList = list(range(self._startJobIndex, sequenceLength + self._startJobIndex))
      if sequenceLength != len(self._jobIndexList):
        return self._reportError("JobIndexList (length=%s) does not have the correct length of %s" %
                                 (len(self._jobIndexList), sequenceLength))
      self.setParameterSequence('JobIndexList', self._jobIndexList, addToWorkflow='JobIndex')
      self._addParameter(self.workflow, 'JobIndex', 'int', self._startJobIndex, 'JobIndex')

    LOG.notice("Job splitting successful")

    return S_OK()

  def _splitByData(self):
    """Split into one job per input file.

    :return: parameter name and parameter values for setParameterSequence()
    :rtype: list(tuple(str, list, bool/str))
    """
    # reset split attribute to avoid infinite loop
    self._splittingOption = None

    LOG.notice("Job splitting: Splitting 'byData' method...")

    # Ensure that data have been specified by setInputData() method
    if not self._data:
      self._reportError("Job splitting: missing input data")
      return []

    if self._numberOfFilesPerJob > len(self._data):
      self._reportError("'numberOfFilesPerJob' must be less/equal than the number of input data")
      return []

    data = breakListIntoChunks(self._data, self._numberOfFilesPerJob)

    LOG.notice('Job splitting: submission consists of %d job(s)' % len(data))

    return [('InputData', data, 'ParametricInputData')]

  def _splitByEvents(self):
    """Split into job per subset of events.

    :return: parameter name and parameter values for setParameterSequence()
    :rtype: list(tuple(str, list, bool/str))
    """
    # reset split attribute to avoid infinite loop
    self._splittingOption = None

    LOG.notice("Job splitting: splitting 'byEvents' method...")

    if self._eventsPerJob and self._numberOfJobs:
      # 1st case: (numberOfJobs=3, eventsPerJob=10)
      # trivial case => each job (total of 3) run applications of 10 events each
      LOG.debug("Job splitting: events per job and number of jobs")

      mapEventJob = [self._eventsPerJob] * self._numberOfJobs

    elif self._eventsPerJob and self._totalNumberOfEvents:
      # 2nd case: (split="byEvents", eventsPerJob=10, totalNumberOfEvents=10)
      # Given the number of events per job and total of number of event we want,
      # we can compute the unknown which is the number of jobs.

      LOG.debug("Job splitting: Events per job and total number of events")

      if self._eventsPerJob > self._totalNumberOfEvents:
        self._reportError("Job splitting: The number of events per job has to be lower than or equal"
                          "to the total number of events")
        return []

      numberOfJobsIntDiv = self._totalNumberOfEvents // self._eventsPerJob
      numberOfJobsRest = self._totalNumberOfEvents % self._eventsPerJob

      mapEventJob = [self._eventsPerJob] * numberOfJobsIntDiv

      mapEventJob += [numberOfJobsRest] if numberOfJobsRest != 0 else []

    else:

      # 3rd case: (split='byEvents', njobs=10, totalNumberOfEvents=10)
      # Then compute the right number of events per job
      LOG.debug("Job splitting: The number of jobs and the total number of events")

      if (not self._totalNumberOfEvents) or (self._totalNumberOfEvents < self._numberOfJobs):
        self._reportError("Job splitting: The number of events has to be greater than or equal to the number of jobs")
        return []

      eventPerJobIntDiv = self._totalNumberOfEvents // self._numberOfJobs
      eventPerJobRest = self._totalNumberOfEvents % self._numberOfJobs

      mapEventJob = [eventPerJobIntDiv] * self._numberOfJobs

      if eventPerJobRest != 0:
        for suplement in range(eventPerJobRest):
          mapEventJob[suplement] += 1

    LOG.debug("Job splitting: events over the jobs: %s" % mapEventJob)
    LOG.notice("Job splitting: submission consists of %d job(s)" % len(mapEventJob))

    return [('NumberOfEvents', mapEventJob, 'NbOfEvts')]

  def _splitBySkip(self):
    """Split the input data over many jobs per file."""
    lfns = self._data
    eventsPerFile = self._eventsPerFile
    eventsPerJob = self._eventsPerJob
    jobsPerFile = eventsPerFile // eventsPerJob
    remainder = eventsPerFile % eventsPerJob
    if remainder > 0:
      jobsPerFile += 1
    # we only need to split for the first application, the rest picks it up from there...
    # we have to set the inputData, startFrom and numberOfEvents parameter for each input data
    inputDataList = []
    startFromList = []
    nEventsList = []
    for lfn in lfns:
      inputDataList.extend([lfn] * jobsPerFile)
      startFromList.extend(list(range(0, eventsPerFile, eventsPerJob)))
      nEventsList.extend([eventsPerJob] * jobsPerFile)
      if remainder > 0:
        nEventsList[-1] = remainder

    self._numberOfJobs = len(inputDataList)

    LOG.notice("Job splitting: submission consists of %d job(s)" % self._numberOfJobs)

    return [('InputData', inputDataList, 'InputData'),
            ('startFrom', startFromList, 'startFrom'),
            ('NumberOfEvents', nEventsList, 'NbOfEvts'),
            ]

  def setSplittingStartIndex(self, start):
    """Set the initial job index for the JobIndex parameter used to define the output file name index.

    :param int start: value where to start from, must be positive integer
    :returns: S_OK, S_ERROR
    """
    start = toInt(start, cond=lambda x: x > 0)
    if not start:
      return self._reportError("Start Index must be positive integer")
    self._startJobIndex = start
    return S_OK()

  def setSplitJobIndexList(self, jobIndexList):
    """Set the list of job indices to use.

    Define the list of indices to use to name output files. By default numbers from *0* to *n* are used.
    This function should be used when rerunning a select sample of input files for example.

    The length of the list must be as long as the number of jobs.

    See also :func:`setSplittingStartIndex`

    :param list jobIndexList: list of job indices
    """
    self._checkArgs({'jobIndexList': list})
    self._jobIndexList = list(jobIndexList)
    return S_OK()

  def setSplitOutputData(self, listOfOutputFiles, OutputPath='', OutputSE=''):
    """Set the list of output file names.

    :param list listOfOutputFiles: list of strings or list of list to set the output files for each split job.
    :param str OutputPath: Optional parameter to specify the Path in the Storage, postpended to /ilc/user/u/username/
    :param OutputSE: Optional parameter to specify the Storage Element to store data or files, e.g. CERN-SRM
    :type OutputSE: `python:list` or `str`
    """
    self._checkArgs({'listOfOutputFiles': list})
    self.setOutputData('PARAMETRIC_OUTPUT', OutputPath, OutputSE)
    self.setParameterSequence('OutputDataSplit', listOfOutputFiles, addToWorkflow='UserOutputData')

  def setSplitParameter(self, parameterName, listOfValues):
    """Set a list of parameter values that can be used to replace a named placeholder in application parameters.

    This only works for application parameters, not for job parameters, and at the moment only for string types.

    For example::

      outputFiles = ['electron.slcio', 'muon.slcio', 'pion.slcio']
      particles = ['e-', 'mu-', 'pi-']
      job = UserJob()
      job.setSplitDoNotAlterOutputFilename()
      job.setName('DDSimTest_%n')
      job.setSplitParameter('particle', particles)
      job.setSplitParameter('outputFile', outputFiles)
      job.setSplitOutputData(outputFiles, 'test/ddsim', 'CERN-DST-EOS')
      job.setOutputSandbox('*.log')
      ddsim = DDSim()
      ddsim.setVersion('ILCSoft-2018-08-10_gcc62')
      ddsim.setDetectorModel('CLIC_o3_v14')

      # the named placeholder '%(particle)s' has the same name as the first argument of setSplitParameter
      ddsim.setExtraCLIArguments('--gun.particle=%(particle)s')
      ddsim.setOutputFile('%(outputFile)s.slcio')

    :param str parameterName: name of the parameter and the placeholder
    :param list listOfValues: the list of values to be used, one per job
    """
    self.setParameterSequence(parameterName + 's', listOfValues, addToWorkflow=parameterName)
    self._addParameter(self.workflow, parameterName, 'string', '', parameterName)
